"""
Malware Detection by Eating a Whole EXE
Edward Raff, Jon Barker, Jared Sylvester, Robert Brandon, Bryan Catanzaro, Charles Nicholas
https://arxiv.org/abs/1710.09435
"""
import os
import torch
import torch.nn as nn
import torch.nn.functional as F

from secml_malware.models.basee2e import End2EndModel

from secml.settings import SECML_PYTORCH_USE_CUDA

use_cuda = torch.cuda.is_available() and SECML_PYTORCH_USE_CUDA


class MalConv(End2EndModel):
	"""
	Architecture implementation.
	"""

	def __init__(self, pretrained_path=None, embedding_size=8, max_input_size=2 ** 20):
		super(MalConv, self).__init__(embedding_size, max_input_size, 256, False)
		self.embedding_1 = nn.Embedding(num_embeddings=257, embedding_dim=embedding_size)
		self.conv1d_1 = nn.Conv1d(in_channels=embedding_size, out_channels=128, kernel_size=(500,), stride=(500,),
								  groups=1, bias=True)
		self.conv1d_2 = nn.Conv1d(in_channels=embedding_size, out_channels=128, kernel_size=(500,), stride=(500,),
								  groups=1, bias=True)
		self.dense_1 = nn.Linear(in_features=128, out_features=128, bias=True)
		self.dense_2 = nn.Linear(in_features=128, out_features=1, bias=True)
		if pretrained_path is not None:
			self.load_simplified_model(pretrained_path)
		if use_cuda:
			self.cuda()

	def embed(self, input_x, transpose=True):
		if isinstance(input_x, torch.Tensor):
			x = input_x.clone().detach().requires_grad_(True).type(torch.LongTensor)
		else:
			x = torch.from_numpy(input_x).type(torch.LongTensor)
		x = x.squeeze(dim=1)
		if use_cuda:
			x = x.cuda()
		emb_x = self.embedding_1(x)
		if transpose:
			emb_x = torch.transpose(emb_x, 1, 2)
		return emb_x

	def embedd_and_forward(self, x):
		conv1d_1 = self.conv1d_1(x)
		conv1d_2 = self.conv1d_2(x)
		conv1d_1_activation = torch.relu(conv1d_1)
		conv1d_2_activation = torch.sigmoid(conv1d_2)
		multiply_1 = conv1d_1_activation * conv1d_2_activation
		global_max_pooling1d_1 = F.max_pool1d(input=multiply_1, kernel_size=multiply_1.size()[2:])
		global_max_pooling1d_1_flatten = global_max_pooling1d_1.view(global_max_pooling1d_1.size(0), -1)
		dense_1 = self.dense_1(global_max_pooling1d_1_flatten)
		dense_1_activation = torch.relu(dense_1)
		dense_2 = self.dense_2(dense_1_activation)
		dense_2_activation = torch.sigmoid(dense_2)
		return dense_2_activation



import numpy as np
import torch
from secml.array import CArray
from secml.settings import SECML_PYTORCH_USE_CUDA

from secml_malware.attack.whitebox import CEnd2EndMalwareEvasion
from secml_malware.models import CClassifierEnd2EndMalware

use_cuda = torch.cuda.is_available() and SECML_PYTORCH_USE_CUDA


def gradient_search(
		start_byte: int,
		gradient: torch.Tensor,
		embedding_bytes: torch.Tensor,
		invalid_val=np.infty,
		invalid_pos: int = -1
):
	"""
	Given the starting byte, the gradient and the embedding map,it returns a list of distances

	Parameters
	----------
	start_byte : int
		the starting byte for the search
	gradient : torch.Tensor
		the gradient
	embedding_bytes : torch.Tensor
		the embedding matrix with all the byte embedded
	invalid_val : optional, default np.infty
		the invalid value to use. Default np.infty
	invalid_pos : int, optional, default -1
		the position of the padding value.
	Returns
	-------

	"""
	if torch.equal(gradient, torch.zeros(gradient.shape)):
		invalid_distances = torch.tensor([invalid_val] * len(embedding_bytes))
		if use_cuda:
			invalid_distances = invalid_distances.cuda()
		return invalid_distances
	distance = torch.zeros(257)
	start_emb_byte = embedding_bytes[start_byte]
	gs = -gradient / torch.norm(gradient)
	if use_cuda:
		distance = distance.cuda()
	for i, b in enumerate(embedding_bytes):
		bts = b - start_emb_byte
		s_i = torch.dot(gs, bts)
		if s_i <= 0:
			distance[i] = invalid_val
		else:
			d_i = torch.norm(b - (start_emb_byte + s_i * gs))
			distance[i] = d_i
	distance[invalid_pos] = invalid_val
	return distance


class CDiscreteBytesEvasion(CEnd2EndMalwareEvasion):
	"""
	Creates the attack that perturbs the header of a Windows PE malware.
	"""

	def __init__(
			self,
			end2end_model: CClassifierEnd2EndMalware,
			index_to_perturb: list,
			iterations: int = 100,
			is_debug: bool = False,
			random_init: bool = False,
			threshold: float = 0.5,
			penalty_regularizer: float = 0,
			chunk_hyper_parameter: int = 256
	):
		"""
		Creates the evasion object

		Parameters
		----------
		end2end_model : CClassifierEnd2EndMalware
			the target end-to-end model
		index_to_perturb : list
			a list containing the index to perturb inside the samples
		iterations : int, optional, default 100
			the number of iterations of the optimizer
		is_debug : bool, optional, default False
			if True, prints debug information during the optimization
		random_init : bool, optional, default False
			if True, it randomizes the locations set by index_to_perturb before starting the optimization
		threshold : float, optional, default 0
			the detection threshold to bypass. Default is 0
		penalty_regularizer : float
			the reularization parameter, Default is 0
		chunk_hyper_parameter : int, optional, default 256
			how many bytes to optimize at each round. Default is 256
		"""
		super(CDiscreteBytesEvasion, self).__init__(
			end2end_model=end2end_model,
			indexes_to_perturb=index_to_perturb,
			iterations=iterations,
			is_debug=is_debug,
			random_init=random_init,
			threshold=threshold,
			penalty_regularizer=penalty_regularizer,
		)
		# if chunk_hyper_parameter is None:
		# 	chunk_hyper_parameter = 256
		self.chunk_hyper_parameter = chunk_hyper_parameter

	def compute_penalty_term(self, original_x: CArray, adv_x: CArray, par: float) -> torch.Tensor:
		"""
		Computes the penalty term as torch node

		Parameters
		----------
		original_x : CArray
			the original malware sample
		adv_x : CArray
			the adversarial malware version
		par : float
			the regularization parameter
		Returns
		-------
		torch.Tensor
			a torch node of the graph containing the penalty
		"""
		emb_original_x = self.classifier.embed(original_x.tondarray())
		emb_adv_x = self.classifier.embed(adv_x.tondarray())
		lambda_variable = torch.autograd.Variable(torch.Tensor([par]), requires_grad=True)
		penalty_term = torch.autograd.Variable(emb_original_x - emb_adv_x, requires_grad=True)
		penalty_term = penalty_term.norm() * lambda_variable
		return penalty_term

	def loss_function_gradient(self, original_x: CArray, adv_x: CArray, penalty_term: torch.Tensor) -> torch.Tensor:
		"""
		Compute the gradient of the loss function of the target model

		Parameters
		----------
		original_x : CArray
			the original malware sample
		adv_x : CArray
			the adversarial malware sample
		penalty_term : torch.Tensor
			the penalty term

		Returns
		-------
		torch.Tensor
			the gradient of the model w.r.t. input on the embedding layer
		"""
		emb_adv_x = self.classifier.embed(adv_x.tondarray())
		if use_cuda:
			penalty_term = penalty_term.cuda()
		y = self.classifier.embedding_predict(emb_adv_x)
		output = y + penalty_term
		if use_cuda:
			output = output.cuda()
		g = torch.autograd.grad(output, emb_adv_x)[0]
		g = torch.transpose(g, 1, 2)[0]
		return g

	def optimization_solver(self, E: torch.Tensor, gradient_f: torch.Tensor, index_to_consider: list,
							x_init: CArray) -> CArray:
		"""
		Optimizes the end-to-end evasion

		Parameters
		----------
		E : torch.Tensor
			the embedding matrix E, with all the embedded values
		gradient_f : torch.Tensor
			the gradient of the function w.r.t. the embedding
		index_to_consider : list
			the list of indexes to perturb
		x_init : CArray
			the input sample to manipulate

		Returns
		-------
		CArray
			the adversarial malware
		"""
		if self.chunk_hyper_parameter:
			best_indexes = gradient_f[index_to_consider].norm(dim=1).argsort(descending=True)[
						   :self.chunk_hyper_parameter]
			best_indexes = [index_to_consider[i] for i in best_indexes]
		else:
			best_indexes = index_to_consider
		results = [self._find_byte_using_gradient(E, gradient_f, i, x_init) for i in best_indexes]
		for res in results:
			if not torch.equal(res[1], self._invalid_value):
				x_init[res[2]] = res[0].item()
		return x_init

	def _find_byte_using_gradient(self, E: torch.Tensor, gradient_f: torch.Tensor, i: int, x_init: CArray) -> (
			int, torch.Tensor, int):
		gradient_f_i = gradient_f[i]
		x_i = x_init[i].tondarray().astype(np.uint16).ravel().item()
		distances = gradient_search(
			x_i,
			gradient_f_i,
			E,
			invalid_val=self._invalid_value,
			invalid_pos=self.invalid_pos,
		)
		min_value, byte_to_choose = torch.min(distances, dim=0, keepdim=True)
		return byte_to_choose, min_value, i

	def infer_step(self, x_init: CArray) -> float:
		"""
		Return prediction w.r.t. the malware class

		Parameters
		----------
		x_init : CArray
			the sample to use for the forward step

		Returns
		-------
		float
			the malware score
		"""
		_, confidence = self.classifier.predict(x_init, return_decision_function=True)
		return confidence[1].item()

	def invert_feature_mapping(self, x: CArray, x_adv: CArray) -> CArray:
		"""
		Invert the feature mapping

		Parameters
		----------
		x : CArray
			the original sample
		x_adv : CArray
			the adversarial sample

		Returns
		-------
		CArray
			the inverted feature mapping of the adv sample
		"""
		return x_adv

	def apply_feature_mapping(self, x: CArray) -> CArray:
		"""
		Applies the feature extraction

		Parameters
		----------
		x : CArray
			the input malware sample

		Returns
		-------
		CArray
			the feature vector
		"""
		return x



from secml_malware.attack.whitebox.c_end2end_evasion import CEnd2EndMalwareEvasion
from secml_malware.attack.whitebox.c_discretized_bytes_evasion import CDiscreteBytesEvasion
from secml_malware.attack.whitebox.c_fast_gradient_sign_evasion import CFastGradientSignMethodEvasion
from secml_malware.attack.whitebox.c_format_exploit_evasion import CFormatExploitEvasion
from secml_malware.attack.whitebox.c_header_evasion import CHeaderEvasion
from secml_malware.attack.whitebox.c_padding_evasion import CPaddingEvasion
from secml_malware.attack.whitebox.c_kreuk_evasion import CKreukEvasion
from secml_malware.attack.whitebox.c_suciu_evasion import CSuciuEvasion
from secml_malware.attack.whitebox.c_padding_evasion import CPaddingEvasion
from secml_malware.attack.whitebox.c_extend_dos_evasion import CExtendDOSEvasion
from secml_malware.attack.whitebox.c_shift_evasion import CContentShiftingEvasion


from secml_malware.attack.whitebox.c_discretized_bytes_evasion import CDiscreteBytesEvasion
from secml_malware.models import CClassifierEnd2EndMalware


class CPaddingEvasion(CDiscreteBytesEvasion):
	"""
	Constructs an attack object that append one byte at time.
	"""

	def __init__(
			self,
			end2end_model: CClassifierEnd2EndMalware,
			how_many: int,
			iterations: int = 100,
			is_debug: bool = False,
			random_init: bool = False,
			threshold: float = 0,
			penalty_regularizer: int = 0
	):
		"""
		Create the padding attack

		Parameters
		----------
		end2end_model : CClassifierEnd2EndMalware
			the target end-to-end model
		how_many : int
			how many padding byte to inject
		iterations : int, optional, default 100
			the number of iterations of the optimizer
		is_debug : bool, optional, default False
			if True, prints debug information during the optimization
		random_init : bool, optional, default False
			if True, it randomizes the locations set by index_to_perturb before starting the optimization
		threshold : float, optional, default 0
			the detection threshold to bypass. Default is 0
		penalty_regularizer : float, optional, default 0
			the reularization parameter, Default is 0
		"""
		super(CPaddingEvasion, self).__init__(
			end2end_model,
			index_to_perturb=[],
			iterations=iterations,
			is_debug=is_debug,
			random_init=random_init,
			threshold=threshold,
			penalty_regularizer=penalty_regularizer
		)
		self.how_many_padding_bytes = how_many

	def _run(self, x0, y0, x_init=None):
		invalid_value = 256 if self.invalid_pos == -1 else self.invalid_pos
		padding_positions = x0.find(x0 == invalid_value)
		if not padding_positions:
			self.indexes_to_perturb = []
		else:
			self.indexes_to_perturb = list(
				range(
					padding_positions[0],
					min(x0.size, padding_positions[0] + self.how_many_padding_bytes),
				)
			)
		return super(CPaddingEvasion, self)._run(x0, y0, x_init=x_init)



import numpy as np

from secml_malware.attack.whitebox.c_kreuk_evasion import CKreukEvasion
from secml_malware.models import CClassifierEnd2EndMalware


class CSuciuEvasion(CKreukEvasion):
	def __init__(
			self,
			end2end_model: CClassifierEnd2EndMalware,
			how_many_padding_bytes: int,
			epsilon: float,
			is_debug: bool = False,
			threshold: float = 0.5,
			compute_slack: bool = True
	):
		"""
		Create the padding attack by Suciu et al. https://arxiv.org/abs/1810.08280

		Parameters
		----------
		end2end_model : CClassifierEnd2EndMalware
			the target end-to-end model
		how_many_padding_bytes: int, optional, default 512
			how many padding bytes
		epsilon : float
			the distortion amount
		is_debug : bool, optional, default False
			if True, prints debug information during the optimization
		threshold : float, optional, default 0
			the detection threshold to bypass. Default is 0
		compute_slack : bool, optional, default True
			if True, uses also the slack space between sections. Default True
		"""
		super(CSuciuEvasion, self).__init__(
			end2end_model=end2end_model,
			how_many_padding_bytes=how_many_padding_bytes,
			epsilon=epsilon,
			iterations=1,
			is_debug=is_debug,
			threshold=threshold,
			p_norm=np.infty,
			compute_slack=compute_slack
		)



import unittest

import torch
from secml.array import CArray

from secml_malware.attack.whitebox.tests.malware_test_base import End2EndBaseTests
from secml_malware.models import MalConv


class MalConvTestSuite(End2EndBaseTests):
	def setUp(self):
		super(MalConvTestSuite, self).setUp()
		self.m = MalConv()

	def test_empty_baseline_is_goodware(self):
		self.m.load_simplified_model(self.ember_path)
		y_pred = self.m(torch.tensor(self.baseline))
		if y_pred.is_cuda:
			y_pred = y_pred.cpu()
		y = int(y_pred.detach().numpy()[0][0] > 0.5)
		self.assertEqual(y, 0)

	def test_malconv_load_ember_no_path(self):
		self.classifier.load_pretrained_model()
		y = self.classifier.predict(
			CArray(torch.tensor(self.baseline)), return_decision_function=False
		)
		self.assertEqual(y, 0)

	def test_empty_baseline_pytorch_classifier_wrapper(self):
		self.classifier.load_pretrained_model(self.ember_path)
		y = self.classifier.predict(
			CArray(torch.tensor(self.baseline)), return_decision_function=False
		)
		self.assertEqual(y, 0)

	def test_batch_prediction_malconv(self):
		self.classifier.load_pretrained_model(self.ember_path)
		y_pred = self.classifier.predict(self.X)

		y_pred1, y_pred2 = self.classifier.predict(self.X[0, :]), self.classifier.predict(self.X[1, :])

		self.assert_array_equal(y_pred1, y_pred[0])
		self.assert_array_equal(y_pred2, y_pred[1])


if __name__ == "__main__":
	unittest.main()



import copy

import lief
from secml.array import CArray

from secml_malware.attack.blackbox.c_blackbox_problem import CBlackBoxProblem
from secml_malware.attack.blackbox.c_wrapper_phi import CWrapperPhi


class CGammaAPIEvasionProblem(CBlackBoxProblem):

	def __init__(self,
				 api_list: list,
				 model_wrapper: CWrapperPhi,
				 population_size: int,
				 penalty_regularizer: float,
				 iterations: int,
				 seed: int = None,
				 is_debug: bool = False):
		"""
		TODO: not yet tested properly

		Manipulate the sample by adding APIs, using the Gamma approach.

		Parameters
		----------
		api_list : list
			A list of couples [dll_name, function_name] of each API that can be included during the optimization
		model_wrapper : CWrapperPhi
			The model under attack
		population_size : int
			How many samples will be generated by the genetic strategy at each round
		penalty_regularizer : float
			The value of the regularization parameter
		iterations: int
			Specifies how many generations will be created by the genetic algorithm.
			The overall number of queries will be (iterations + 1) * population_size
		seed : int
			Specify a seed for init the random engine
		is_debug : bool
			Prints additional information during the optimization

		Returns
		----------
		An evasion object ready to run
		"""
		super(CGammaAPIEvasionProblem, self).__init__(model_wrapper,
													  len(api_list),
													  population_size,
													  penalty_regularizer,
													  iterations,
													  seed, is_debug)

		self.api_list = api_list
		self._cache_lib = None
		self._names = None

	def _add_import(self, liefpe, dll, func_name):
		lib = [l for l in liefpe.imports if l.name.lower() == dll.lower()]
		if lib == []:
			lib = liefpe.add_library(dll)
		lib = lief.add_library(dll) if lib == [] else lib[0]
		names = set([e.name for e in lib.entries])
		if not func_name in names:
			lib.add_entry(func_name)

	def apply_feasible_manipulations(self, t, x: CArray) -> CArray:
		x_adv = copy.deepcopy(x)
		x_adv = x_adv.tolist()
		lief_adv: lief.PE.Binary = lief.PE.parse(x_adv[0])
		for i in range(t.shape[-1]):
			dll_name, function_name = self.api_list[i]
			self._add_import(lief_adv, dll_name, function_name)
		builder = lief.PE.Builder(lief_adv)
		builder.build_imports(True).patch_imports(True)
		builder.build()
		x_adv = CArray(builder.get_build())
		x_adv = x_adv.reshape((1, x_adv.shape[-1]))
		return x_adv



import copy
from secml.array import CArray

from secml_malware.attack.blackbox.c_gamma_evasion import CGammaEvasionProblem
from secml_malware.attack.blackbox.c_wrapper_phi import CWrapperPhi
from secml_malware.utils.pe_operations import create_int_list_from_x_adv, shift_section_by


class CGammaShiftEvasionProblem(CGammaEvasionProblem):
	def __init__(self,
				 section_population: list,
				 model_wrapper: CWrapperPhi,
				 population_size: int,
				 penalty_regularizer: float,
				 iterations: int,
				 seed: int = None,
				 is_debug: bool = False,
				 hard_label : bool = False,
				 threshold : float = 0.5,
				 loss : str = 'l1'
				 ):
		"""
		TODO: not yet tested properly
		Creates the GAMMA section injection attack

		Parameters
		----------
		section_population : list
			a list containing all the goodware sections to inject
		model_wrapper : CWrapperPhi
			the target model, wrapped inside a CWrapperPhi
		population_size : int
			the population size generated at each round by the genetic algorithm
		penalty_regularizer: float
			the regularization parameter used for the size constraint
		iterations : int, optional, default 100
			the total number of iterations, default 100
		seed : int, optional, default None
			specifies an initialization seed for the random. None for not using determinism
		is_debug : bool, optional, default False
			if True, it prints messages while optimizing. Default is False
		hard_label : bool, optional default False
			if True, the problem will use only binary labels instead. Infinity will be used for non-evasive samples.
		threshold : float, optional, default 0
			the detection threshold. Leave 0 to test the degradation of the model until the end of the algorithm.
		loss : str, optional, default l1
			The loss function used as objective function
		"""
		super(CGammaShiftEvasionProblem, self).__init__(section_population,
														model_wrapper,
														population_size,
														penalty_regularizer,
														iterations,
														seed,
														is_debug,
														hard_label,
														threshold,
														loss)


	def apply_feasible_manipulations(self, t, x: CArray) -> CArray:
		x_adv = copy.deepcopy(x)
		overall_content = []
		for i in range(t.shape[-1]):
			content = self.section_population[i]
			content_to_append = content[:int(round(len(content) * t[i]))]
			overall_content.extend(content_to_append)
		overall_size = len(overall_content)
		x_adv, indexes_to_perturb = self._craft_perturbed_c_array(x_adv, overall_size)
		x_adv[0, indexes_to_perturb[:overall_size]] = CArray(overall_content)
		x_adv = x_adv.reshape((1, x_adv.shape[-1]))
		return x_adv

	def _craft_perturbed_c_array(self, x0 : CArray, size : int):
		x_init, indexes_to_perturb = self._generate_list_adv_example(x0, size)
		self.indexes_to_perturb = indexes_to_perturb
		x_init = CArray([x_init])
		return x_init, indexes_to_perturb

	def _generate_list_adv_example(self, x0, size):
		x_init = create_int_list_from_x_adv(x0, 256, False)
		x_init, index_to_perturb_sections = shift_section_by(x_init, preferable_extension_amount=size)
		return x_init, index_to_perturb_sections


import copy
import random
import time
from typing import Tuple, Any

import numpy as np
from secml.adv.attacks import CAttackEvasion
from secml.array import CArray
from secml.data import CDataset

from secml_malware.attack.blackbox.c_blackbox_problem import CBlackBoxProblem

try:
	import deap
	from deap import base, creator, tools, algorithms
except ImportError:
	raise ImportError(
		"Install DEAP to apply black box evasion")


def _pad_sequence_with_last(sequence, until):
	how_many = until - len(sequence)
	if how_many <= 0:
		return sequence
	return sequence + [sequence[-1]] * how_many


class CGeneticAlgorithm(CAttackEvasion):
	"""
	The genetic optimizer.
	"""

	def _objective_function_gradient(self, x):
		raise NotImplementedError("This function is not used inside secML malware")

	def _objective_function(self, x):
		raise NotImplementedError("This function is not used inside secML malware")

	def objective_function_gradient(self, x):
		raise NotImplementedError("This function is not used inside secML malware")

	def objective_function(self, x):
		raise NotImplementedError("This function is not used inside secML malware")

	def f_eval(self):
		raise NotImplementedError("This function is not used inside secML malware")

	def grad_eval(self):
		raise NotImplementedError("This function is not used inside secML malware")

	def __init__(
			self,
			problem: CBlackBoxProblem,
			is_debug: bool = False,
	):
		"""
		Create and instance of the genetic optimizer.

		Parameters
		----------
		problem : CBlackBoxProblem
			The problem to optimize
		is_debug : bool, default False
			If True, debug prints will be displayed during the attack.
			Default is False
		"""
		CAttackEvasion.__init__(
			self,
			problem.model_wrapper.classifier,
			problem.model_wrapper.classifier,
		)
		self.problem = problem
		self.confidences_ = []
		self.fitness_ = []
		self.sizes_ = []
		self.changes_per_iterations_ = []
		self.model_wrapper = problem.model_wrapper
		self.is_debug = is_debug
		self._original_x = None
		self.minimization_result_ = []
		self.evolved_problem_ = None
		self.stagnation = 5
		self.elapsed_time_ = 0

	def run(self, x, y, ds_init=None) -> Tuple[CArray, CArray, CDataset, Any]:
		"""
		Runs the genetic algorithms.

		Parameters
		----------
		x : CArray
			input sample to perturb
		y : CArray
			original class
		ds_init : CDataset, optional, default None
			the initialization point.
			Default is None
		Returns
		-------
		CArray
			y_pred : the predicted label after the attack
		CArray
			scores : the scores after the attack
		CDataset
			adv_ds : the CDataset containing the adversarial points
		CArray
			f_obj : the mean value for the objective function
		"""
		x = CArray(x).atleast_2d()
		y = CArray(y).atleast_2d()
		x_init = None if ds_init is None else CArray(ds_init.X).atleast_2d()

		# only consider samples that can be manipulated
		v = self.is_attack_class(y)
		idx = CArray(v.find(v)).ravel()
		# print(v, idx)

		# number of modifiable samples
		n_mod_samples = idx.size

		adv_ds = CDataset(x.deepcopy(), y.deepcopy())

		# If dataset is sparse, set the proper attribute
		if x.issparse is True:
			self._issparse = True

		# array in which the value of the optimization function are stored
		fs_opt = CArray.zeros(n_mod_samples, )
		y_pred = CArray.zeros(n_mod_samples, )
		scores = CArray.zeros((n_mod_samples, 2))
		for i in range(n_mod_samples):
			k = idx[i].item()  # idx of sample that can be modified

			xi = x[k, :] if x_init is None else x_init[k, :]
			x_opt, f_opt = self._run(x[k, :], y[k], x_init=xi)

			self.logger.info(
				"Point: {:}/{:}, f(x):{:}, eval:{:}/{:}".format(k, x.shape[0], f_opt, self.f_eval, self.grad_eval))
			if x_opt.shape[-1] > adv_ds.X.shape[-1]:
				# Need to resize the whole adv dataset, since CDataset can't deal with varying vector sizes
				new_length = x_opt.shape[-1]
				adv_ds.X = adv_ds.X.resize((adv_ds.X.shape[0], new_length), 256)
			adv_ds.X[k, :min(adv_ds.X.shape[-1], x_opt.shape[-1])] = x_opt
			fs_opt[i] = f_opt
			y_p, score = self.problem.model_wrapper.predict(x_opt, return_decision_function=True)
			scores[i, :] = score[0, :]
			y_pred[i] = y_p

		# Return the mean objective function value on the evasion points (
		# computed from the outputs of the surrogate classifier)
		f_obj = fs_opt.mean()

		return y_pred, scores, adv_ds, f_obj

	def _run(self, x0, y0, x_init=None):

		if x_init is None:
			x_init = copy.deepcopy(x0)

		self._original_x = self.problem.init_starting_point(x_init)
		current_conf = self.problem.score_step(x_init, 0)

		if self.is_debug:
			print(f'> Original Confidence: {self.confidences_[0]}')
			print("> Beginning new sample evasion...")

		minimization_results = self._compute_black_box_optimization()

		self.minimization_result_ = minimization_results.tondarray()

		x_adv = self.problem.apply_feasible_manipulations(self.minimization_result_, self._original_x)

		if self.is_debug:
			print(f'>AFTER INVERSION, CONFIDENCE SCORE: {self.confidences_[-1]}')
		return x_adv, self.confidences_[-1]

	def _compute_black_box_optimization(self) -> CArray:

		start_t = time.time()

		slice_indexes = [self.problem.population_size]

		creator.create("FitnessMin", base.Fitness, weights=(-1.0,))
		creator.create("Individual", list, fitness=creator.FitnessMin)

		toolbox = base.Toolbox()
		toolbox.register('attr_float', random.random)
		toolbox.register("individual", tools.initRepeat, creator.Individual, toolbox.attr_float,
						 n=self.problem.latent_space_size)
		toolbox.register("population", tools.initRepeat, list, toolbox.individual)

		toolbox.register("evaluate", self.problem.fitness)
		toolbox.register("mate", tools.cxOnePoint)
		toolbox.register("mutate", random_mutation, indpb=0.3)
		toolbox.register("select", tools.selTournament, tournsize=self.problem.population_size)

		pop = toolbox.population(n=self.problem.population_size)
		fitness = [self.problem.fitness(np.array(t)) for t in pop]
		for ind, fit in zip(pop, fitness):
			ind.fitness.values = fit

		# CXPB  is the probability with which two individuals are crossed
		# MUTPB is the probability for mutating an individual
		CXPB, MUTPB = 0.9, 0.3
		g = 0
		last_n_best_fits = []
		while g < self.problem.iterations:
			# Select the next generation individuals
			offspring = toolbox.select(pop, self.problem.population_size)
			# Clone the selected individuals
			offspring = list(map(toolbox.clone, offspring))

			# Apply crossover and mutation on the offspring
			for child1, child2 in zip(offspring[::2], offspring[1::2]):
				if random.random() < CXPB:
					toolbox.mate(child1, child2)
					del child1.fitness.values
					del child2.fitness.values

			for mutant in offspring:
				if random.random() < MUTPB:
					toolbox.mutate(mutant)
					del mutant.fitness.values

			# Evaluate the individuals with an invalid fitness
			invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
			slice_indexes.append(len(invalid_ind))
			fitness = [self.problem.fitness(np.array(t)) for t in invalid_ind]
			for ind, fit in zip(invalid_ind, fitness):
				ind.fitness.values = fit

			pop.extend(invalid_ind)
			fits = [ind.fitness.values[0] for ind in pop]
			best_fitness = min(fits)
			last_n_best_fits.insert(0, best_fitness)
			last_n_best_fits = last_n_best_fits[:self.stagnation]
			if self.is_debug:
				print(f'>{g} - Global min: {best_fitness}')
			if len(last_n_best_fits) == self.stagnation and (
					all((np.array(last_n_best_fits) - best_fitness) < 1e-6) or all(
					np.array(last_n_best_fits) == np.infty)):
				if self.is_debug:
					print('Stagnating result!')
				break
			g += 1

		confidences, fitness, sizes = self.problem._export_internal_results(slice_indexes)
		end_t = time.time()
		self.confidences_ = _pad_sequence_with_last(confidences, self.problem.iterations)
		self.fitness_ = _pad_sequence_with_last(fitness, self.problem.iterations)
		self.sizes_ = _pad_sequence_with_last(sizes, self.problem.iterations)
		self.elapsed_time_ = end_t - start_t
		best_t = tools.selBest(pop, 1)[0]

		del creator.FitnessMin
		del creator.Individual

		return CArray(best_t)

	@classmethod
	def write_adv_to_file(cls, x_adv: CArray, path: str):
		"""
		Write the adversarial malware as a file on disk

		Parameters
		----------
		x_adv : CArray
			The adversarial malware to dump
		path : str
			The path where to save the executable
		"""
		x_real = x_adv.tolist()[0]
		x_real_adv = b''.join([bytes([i]) for i in x_real])
		with open(path, 'wb') as f:
			f.write(x_real_adv)


def random_mutation(individual, indpb):
	"""
	Apply the mutation operator, that perturb randomly each entry of the individual, with a given probability.
	The mutation is applied in-place.

	Parameters
	----------
	individual :
		the individual to mutate
	indpb : float
		the probability of altering a single entry
	Returns
	-------
	tuple
		the mutated individual, the mutatio is in-place
	"""
	size = len(individual)
	for i in range(size):
		if random.random() < indpb:
			individual[i] = random.random()
	return individual,



import copy
import random
import string
import numpy as np
import lief
from secml.array import CArray

from secml_malware.attack.blackbox.c_blackbox_problem import _slice_sequence
from secml_malware.attack.blackbox.c_gamma_evasion import CGammaEvasionProblem
from secml_malware.attack.blackbox.c_wrapper_phi import CWrapperPhi


class CGammaSectionsEvasionProblem(CGammaEvasionProblem):
	"""
	GAMMA section injection attack class
	"""
	def __init__(self,
				 section_population: list,
				 model_wrapper: CWrapperPhi,
				 population_size: int,
				 penalty_regularizer: float,
				 iterations: int,
				 seed: int = None,
				 is_debug: bool = False,
				 hard_label : bool = False,
				 threshold : float = 0.5,
				 loss : str = 'l1',
				 random_names : bool = True
				 ):
		"""
		Creates the GAMMA section injection attack

		Parameters
		----------
		section_population : list
			a list containing all the goodware sections to inject
		model_wrapper : CWrapperPhi
			the target model, wrapped inside a CWrapperPhi
		population_size : int
			the population size generated at each round by the genetic algorithm
		penalty_regularizer: float
			the regularization parameter used for the size constraint
		iterations : int, optional, default 100
			the total number of iterations, default 100
		seed : int, optional, default None
			specifies an initialization seed for the random. None for not using determinism
		is_debug : bool, optional, default False
			if True, it prints messages while optimizing. Default is False
		hard_label : bool, optional default False
			if True, the problem will use only binary labels instead. Infinity will be used for non-evasive samples.
		threshold : float, optional, default 0
			the detection threshold. Leave 0 to test the degradation of the model until the end of the algorithm.
		loss : str, optional, default l1
			The loss function used as objective function
		random_names: bool
			Uses random names for the new sections. Default True
		"""
		super(CGammaSectionsEvasionProblem, self).__init__(section_population,
														model_wrapper,
														population_size,
														penalty_regularizer,
														iterations,
														seed,
														is_debug,
														hard_label,
														threshold,
														loss)
		self.random_names = random_names
		self.names_ = []
		self.best_names_ = []

	def init_starting_point(self, x: CArray) -> CArray:
		"""
		Initialize the problem, by setting the starting point.

		Parameters
		----------
		x : CArray
			the initial point

		Returns
		-------
		CArray
			the initial point (padded accordingly to remove trailing invalid values)
		"""
		original = super().init_starting_point(x)
		self.best_names_ = []
		self.names_ = []
		return original

	def apply_feasible_manipulations(self, t, x: CArray) -> CArray:
		"""
		Applies the section injection manipulation.

		Parameters
		----------
		t : np.ndarray
			the vector of parameters specifying how much content must be included
		x : CArray
			the original malware

		Returns
		-------
		CArray
			the adversarial malware
		"""
		x_adv = copy.deepcopy(x)
		x_adv = x_adv[0,:].flatten().tolist()
		lief_adv: lief.PE.Binary = lief.PE.parse(raw=x_adv)
		names = []
		for i in range(self.latent_space_size):
			content = self.section_population[i]
			content_to_append = content[:int(round(len(content) * t[i]))]
			if self.best_names_ != []:
				section_name = self.best_names_[i]
			else:
				section_name = ''.join(random.choice(string.ascii_letters) for _ in range(8))
			names.append(section_name)
			s = lief.PE.Section(section_name)
			s.content = content_to_append
			lief_adv.add_section(s)
		self.names_.append(names)
		builder = lief.PE.Builder(lief_adv)
		builder.build()
		x_adv = CArray(builder.get_build())
		x_adv = x_adv.reshape((1, x_adv.shape[-1]))
		return x_adv

	def _export_internal_results(self, irregular=None):
		confidence, fitness, sizes = super()._export_internal_results(irregular)
		index = int(np.argmin(self.fitness_[1:]))
		self.best_names_ = self.names_[index]
		return confidence, fitness, sizes



import itertools
import math
import struct

import lief
import numpy as np
from secml.array import CArray


def shift_pointer_to_section_content(liefpe: lief.PE.Binary, raw_code: bytearray, entry_index: int, amount: int,
									 pe_shifted_by: int = 0) -> bytearray:
	"""
	Shifts the section content pointer.

	Parameters
	----------
	liefpe : lief.PE.Binary
		the binary wrapper by lief
	raw_code : bytearray
		the code of the executable to eprturb
	entry_index : int
		the entry of the section to manipulate
	amount : int
		the shift amount
	pe_shifted_by : int, optional, default 0
		if the PE header was shifted, this value should be set to that amount
	Returns
	-------
	bytearray
		the modified code
	"""
	pe_position = liefpe.dos_header.addressof_new_exeheader + pe_shifted_by
	optional_header_size = liefpe.header.sizeof_optional_header
	coff_header_size = 24
	section_entry_length = 40
	size_of_raw_data_pointer = 20
	shift_position = (
			pe_position
			+ coff_header_size
			+ optional_header_size
			+ (entry_index * section_entry_length)
			+ size_of_raw_data_pointer
	)
	old_value = struct.unpack("<I", raw_code[shift_position: shift_position + 4])[0]
	new_value = old_value + amount
	new_value = struct.pack("<I", new_value)
	raw_code[shift_position: shift_position + 4] = new_value

	return raw_code


def shift_pe_header(liefpe: lief.PE.Binary, raw_code: bytearray, amount: int) -> bytearray:
	"""
	Shifts the PE header, injecting a default pattern

	Parameters
	----------
	liefpe : lief.PE.Binary
		the binary wrapper by lief
	raw_code : bytearray
		the code of the executable to perturb
	amount : int
		how much to inject

	Returns
	-------
	bytearray
		the modified code
	"""
	if amount == 0:
		return raw_code
	pe_position = liefpe.dos_header.addressof_new_exeheader
	section_alignment = liefpe.optional_header.section_alignment
	file_alignment = liefpe.optional_header.file_alignment
	raw_code[0x3C:0x40] = struct.pack("<I", pe_position + amount)

	raw_code[pe_position + 60 + 20 + 4: pe_position + 60 + 20 + 4 + 4] = struct.pack("<I",
																					 liefpe.optional_header.sizeof_headers + amount)
	pattern = itertools.cycle("I love ToucanStrike <3")
	[raw_code.insert(pe_position, ord(next(pattern))) for _ in range(amount)]

	return raw_code


def apply_shift(file_name: str, new_file_name: str = None, amount: int = 0x200) -> bytearray:
	"""
	Applies the content shifting manipulations to the sample pointed by the path

	Parameters
	----------
	file_name : str
		the file path
	new_file_name : str, optional, default None
		path where to save perturbed sample, if not None
	amount : int, optional, default 512
		the amount to inject. Default is 512
	Returns
	-------
	bytearray
		the perturbed code
	"""
	file_path = file_name
	with open(file_path, "rb") as f:
		code = bytearray(f.read())
	return apply_shift_to_raw_code(amount, code, new_file_name)


def shift_pe_header_by(x: list, preferable_extension_amount: int) -> (list, list):
	"""
	Applies the DOS header extension to a sample contained inside a list

	Parameters
	----------
	x : list
		the sample as a list of integers
	preferable_extension_amount : int
		how much extension

	Returns
	-------
	list, list
		returns the perturbed sample and which are the indexes that can be perturbed
	"""
	if preferable_extension_amount == 0:
		return x, []
	liefpe = lief.PE.parse(x)
	section_file_alignment = liefpe.optional_header.file_alignment
	if section_file_alignment == 0:
		return x, []
	first_content_offset = liefpe.dos_header.addressof_new_exeheader
	extension_amount = int(math.ceil(preferable_extension_amount / section_file_alignment)) * section_file_alignment
	index_to_perturb = list(range(2, 0x3c)) + list(range(0x40, first_content_offset + extension_amount))
	x = shift_pe_header(liefpe, x, extension_amount)
	for i, _ in enumerate(liefpe.sections):
		x = shift_pointer_to_section_content(liefpe, bytearray(x), i, extension_amount, extension_amount)
	return x, index_to_perturb


def shift_section_by(x: list, preferable_extension_amount: int, pe_shifted_by: int = 0) -> (list, list):
	"""
	Applies the content shifting to a sample contained inside a list

	Parameters
	----------
	x : list
		the sample as a list of integers
	preferable_extension_amount : int
		how much extension
	pe_shifted_by : int, optional, default 0
		if the PE header was shifted, this value should be set to that amount

	Returns
	-------
	list, list
		returns the perturbed sample and which are the indexes that can be perturbed
	"""
	if not preferable_extension_amount:
		return x, []
	liefpe = lief.PE.parse(x)
	section_file_alignment = liefpe.optional_header.file_alignment
	if section_file_alignment == 0:
		return x, []
	first_content_offset = liefpe.sections[0].offset
	extension_amount = int(math.ceil(preferable_extension_amount / section_file_alignment)) * section_file_alignment
	index_to_perturb = list(range(first_content_offset, first_content_offset + extension_amount))
	for i, _ in enumerate(liefpe.sections):
		x = shift_pointer_to_section_content(liefpe, x, i, extension_amount, pe_shifted_by)
	x = x[:first_content_offset] + b'\x00' * extension_amount + x[first_content_offset:]
	return x, index_to_perturb


def shift_section_by_using_lief(x: list, liefpe: lief.PE.Binary, preferable_extension_amount: int,
								pe_shifted_by: int = 0) -> (list, list):
	if not preferable_extension_amount:
		return x, []
	section_file_alignment = liefpe.optional_header.file_alignment
	first_content_offset = liefpe.sections[0].offset
	extension_amount = int(math.ceil(preferable_extension_amount / section_file_alignment)) * section_file_alignment
	index_to_perturb = list(range(first_content_offset, first_content_offset + extension_amount))

	# shift offset of each section entry by specified amount
	for i, _ in enumerate(liefpe.sections):
		x = shift_pointer_to_section_content(liefpe, x, i, extension_amount, pe_shifted_by)

	x = x[:first_content_offset] + b'\x00' * extension_amount + x[first_content_offset:]
	return x, index_to_perturb


def apply_shift_to_raw_code(amount: int, code: bytearray, new_file_name: str) -> bytearray:
	"""
	Applies the content shifting manipulation to the sample as bytearray

	Parameters
	----------
	amount : int
		the amount to inject
	code : bytearray
		the code to perturb
	new_file_name : str
		the path where to save the sample. Pass None to skip this.

	Returns
	-------
	bytearray
		the perturbed code
	"""
	parse_pe = lief.PE.parse(list(code))
	amount = (
		parse_pe.optional_header.file_alignment if amount is None else amount
	)
	for i, s in enumerate(parse_pe.sections):
		print(f"Shifting {s.name}")
		code = shift_pointer_to_section_content(parse_pe, code, i, amount)
	code = shift_pe_header(parse_pe, code, amount)
	if new_file_name is not None:
		with open(new_file_name, "wb") as f:
			f.write(code)
		print(f"Written {new_file_name}")
	return code


def create_int_list_from_x_adv(x_adv: CArray, embedding_value: int, is_shifting_values: bool) -> bytearray:
	"""
	Convert CArray sample to list of integers

	Parameters
	----------
	x_adv : CArray
		the sample as a CArray
	embedding_value : int
		the value used for padding the sample
	is_shifting_values : bool
		True if the values are shifted by one

	Returns
	-------
	list
		the sample as list of int
	"""
	invalid_value = 256 if embedding_value == -1 else embedding_value
	padding_positions = x_adv.find(x_adv == invalid_value)
	if padding_positions:
		x_adv = x_adv[:padding_positions[0]]
	if is_shifting_values:
		x_adv = x_adv - 1
	x_adv_edit = x_adv[0, :].astype(np.uint8).flatten().tolist()
	return bytearray(x_adv_edit)


def create_indexes_for_header_fields_manipulations(pe_index: int):
	# COFF manipulations
	# 4 bytes timestamp at PE + 4 (PE size) + 4
	# 4 bytes PointerToSymbolTable at PE + 4 (PE size) + 8
	# 4 bytes NumberOFSymbols at PE + 4 (PE size) + 12
	#
	# Optional header manipulations
	# 1 byte MajorLinkerVersion at PE + 4 + 20 (COFF size) + 2
	# 1 byte MinorLinkerVersion at PE + 4 + 20 (COFF size) + 3
	# 2 byte MajorOperatingSystemVersion at PE + 4 + 20 (COFF size) + 40
	# 2 byte MinorOperatingSystemVersion at PE + 4 + 20 (COFF size) + 42
	# 2 byte MajorImageVersion at PE + 4 + 20 (COFF size) + 44
	# 2 byte MinorImageVersion at PE + 4 + 20 (COFF size) + 46
	
	coff_indexes = list(range(pe_index + 4 + 4, pe_index + 4 + 12))
	optional_header_indexes = [pe_index + 24 + 2, pe_index + 24 + 2]
	optional_header_indexes.extend(list(range(pe_index + 24 + 40, pe_index + 24 + 48)))
	indexes = coff_indexes + optional_header_indexes
	return indexes



import copy
from abc import abstractmethod

import numpy as np
import torch
from secml.adv.attacks import CAttackEvasion
from secml.array import CArray
from secml.array.c_dense import CDense
from secml.settings import SECML_PYTORCH_USE_CUDA

from secml_malware.models import CClassifierEnd2EndMalware

use_cuda = torch.cuda.is_available() and SECML_PYTORCH_USE_CUDA


class CEnd2EndMalwareEvasion(CAttackEvasion):
	"""
	Base abstract class for implementing end-to-end evasion attacks against malware detectors.
	"""

	def _objective_function_gradient(self, x):
		pass

	def _objective_function(self, x):
		pass

	def __init__(
			self,
			end2end_model: CClassifierEnd2EndMalware,
			indexes_to_perturb: list,
			iterations: int = 100,
			is_debug: bool = False,
			random_init: bool = False,
			threshold: float = 0.5,
			penalty_regularizer: float = 0,
			store_checkpoints: int = None

	):
		CAttackEvasion.__init__(
			self,
			end2end_model,
			end2end_model,
		)
		self.iterations = iterations
		self.is_debug = is_debug
		self.indexes_to_perturb = indexes_to_perturb
		self.confidences_ = []
		self.changes_per_iterations_ = []

		self.random_init = random_init

		self.embedding_size = end2end_model.get_embedding_size()
		self.max_input_length = end2end_model.get_input_max_length()
		self.invalid_pos = end2end_model.get_embedding_value()
		self.embedding_value = end2end_model.get_embedding_value()
		self.shift_values = end2end_model.get_is_shifting_values()

		self._invalid_value = torch.tensor([np.infty])

		self.threshold = threshold
		self.penalty_regularizer = penalty_regularizer

		self.store_checkpoints = store_checkpoints

	def f_eval(self):
		pass

	def grad_eval(self):
		pass

	def objective_function_gradient(self, x):
		pass

	def objective_function(self, x):
		pass

	@abstractmethod
	def loss_function_gradient(self, original_x: CArray, adv_x: CArray, penalty_term: torch.Tensor):
		raise NotImplementedError("This is only an abstract method")

	def _run(self, x0, y0, x_init=None):
		"""
		Tries to achieve evasion against and end-to-end classifier.

		Parameters
		----------
		x0 : CArray
			Initial sample.
		y0 : int or CArray
			The true label of x0.
		x_init : CArray or None, optional
			Initialization point. If None, it is set to x0.
		Returns
		-------
		x_opt : CArray
			Evasion sample
		f_opt : float
			Value of objective function on x_opt.
		Notes
		-----
		Internally, it stores the confidences at each round.
		"""

		if x_init is None:
			x_init = copy.copy(x0)

		_, current_conf = self.classifier.predict(x_init, return_decision_function=True)
		current_conf = current_conf[1].item()
		self.confidences_ = [current_conf]

		if self.is_debug:
			print(f'> Original Confidence: {current_conf}')

		if use_cuda:
			self._invalid_value = self._invalid_value.cuda()

		E = self._get_embedded_byte_matrix()
		if self.random_init:
			x_init = self._randomize_values_for_attack(x_init)

		if self.is_debug:
			print("> Beginning new sample evasion...")

		index_to_consider = np.array(self.indexes_to_perturb)
		x_init = self.apply_feature_mapping(x_init)
		for t in range(self.iterations):
			if current_conf < self.threshold:
				if self.is_debug:
					print(f"Stopped at confidence below threshold: {current_conf}/{self.threshold}")
				break

			penalty_term = self.compute_penalty_term(x0, x_init, self.penalty_regularizer)
			gradient_f = self.loss_function_gradient(x0, x_init, penalty_term)
			x_init = self.optimization_solver(E, gradient_f, index_to_consider, x_init)
			current_conf = self.infer_step(x_init)
			if self.store_checkpoints:
				if not t % self.store_checkpoints:
					x_temp = self.invert_feature_mapping(x0, x_init)
					_, check_conf = self.classifier.predict(x_temp, return_decision_function=True)
					check_conf = check_conf[1].item()
					if self.is_debug:
						print(f">{t}/{self.iterations} storing checkpoint: {check_conf}")
					self.confidences_.append(check_conf)
			else:
				self.confidences_.append(current_conf)
			if self.is_debug:
				print(f">{t}/{self.iterations} Shifted confidence:\t{current_conf}")

		x_init = self.invert_feature_mapping(x0, x_init)
		_, current_conf = self.classifier.predict(x_init, return_decision_function=True)
		current_conf = current_conf[1].item()
		if self.is_debug:
			print(f'>AFTER INVERSION, CONFIDENCE SCORE: {current_conf}')
		return x_init, current_conf

	def _get_embedded_byte_matrix(self):
		return self.classifier.embed(
			np.array([[i if i < 256 else 256 for i in range(self.max_input_length)]]),
			transpose=False,
		)[0][:257]

	@abstractmethod
	def apply_feature_mapping(self, x) -> CArray:
		raise NotImplementedError("This method is abstract, you should implement it somewhere else!")

	@abstractmethod
	def invert_feature_mapping(self, x, x_adv) -> CArray:
		raise NotImplementedError("This method is abstract, you should implement it somewhere else!")

	@abstractmethod
	def infer_step(self, x_init) -> CArray:
		raise NotImplementedError("This method is abstract, you should implement it somewhere else!")

	def _randomize_values_for_attack(self, x_init):
		min_val = 0 + self.shift_values
		max_val = 255 + self.shift_values
		x_init[self.indexes_to_perturb] = CDense(
			torch.randint(
				low=min_val, high=max_val, size=(1, len(self.indexes_to_perturb))
			)
		)
		return x_init

	@abstractmethod
	def compute_penalty_term(self, original_x: CArray, adv_x: CArray, par: float) -> float:
		raise NotImplementedError("This method is abstract, you should implement it somewhere else!")

	@abstractmethod
	def optimization_solver(self, E, gradient_f, index_to_consider, x_init) -> CArray:
		raise NotImplementedError("This method is abstract, you should implement it somewhere else!")

	def create_real_sample_from_adv(self, original_file_path: str, x_adv: CArray,
									new_file_path: str = None) -> bytearray:
		"""
		Create a real adversarial example

		Parameters
		----------
		original_file_path : str
			the original malware sample
		x_adv : CArray
			the perturbed malware sample, as created by the optimizer
		new_file_path : str, optional, default None
			the path where to save the adversarial malware. Leave None to not save the result to disk

		Returns
		-------
		bytearray
			the adversarial malware, as string of bytes
		"""
		with open(original_file_path, 'rb') as f:
			code = bytearray(f.read())
		padding_index = x_adv.find(x_adv == self.classifier.get_embedding_value())
		padded_x_adv = copy.copy(x_adv)
		if padding_index:
			padded_x_adv = padded_x_adv[0, :padding_index[0]]
		if self.shift_values:
			padded_x_adv = padded_x_adv - 1
		padded_x_adv = padded_x_adv.astype(np.uint8).flatten().tolist()
		padded_x_adv = b''.join([bytes([i]) for i in padded_x_adv])
		code[:len(padded_x_adv)] = padded_x_adv
		if new_file_path:
			with open(new_file_path, 'wb') as f:
				f.write(code)
		return code



import lightgbm
import numpy as np
from ember import PEFeatureExtractor
from secml.array import CArray
from secml.ml.classifiers import CClassifier


class CClassifierEmber(CClassifier):
	"""
	The wrapper for the EMBER GBDT, by Anderson et al. https://arxiv.org/abs/1804.04637
	"""

	def __init__(self, tree_path: str = None):
		"""
		Create the EMBER tree.

		Parameters
		----------
		tree_path : str
			path to the tree parameters
		"""
		super(CClassifierEmber, self).__init__()
		self._lightgbm_model = self._load_tree(tree_path)

	def extract_features(self, x: CArray) -> CArray:
		"""
		Extract EMBER features

		Parameters
		----------
		x : CArray
			program sample
		Returns
		-------
		CArray
			EMBER features
		"""
		extractor = PEFeatureExtractor(2, print_feature_warning=False)
		x = x.atleast_2d()
		size = x.shape[0]
		features = []
		for i in range(size):
			x_i = x[i, :]
			x_bytes = bytes(x_i.astype(np.int).tolist()[0])
			features.append(np.array(extractor.feature_vector(x_bytes), dtype=np.float32))
		features = CArray(features)
		return features

	def _backward(self, w):
		pass

	def _fit(self, x, y):
		raise NotImplementedError("Fit is not implemented.")

	def _load_tree(self, tree_path):
		booster = lightgbm.Booster(model_file=tree_path)
		self._classes = 2
		self._n_features = booster.num_feature()
		return booster

	def _forward(self, x):
		x = x.atleast_2d()
		scores = self._lightgbm_model.predict(x.tondarray())
		confidence = [[1 - c, c] for c in scores]
		confidence = CArray(confidence)
		return confidence



from secml_malware.attack.whitebox import CFormatExploitEvasion
from secml_malware.models import CClassifierEnd2EndMalware


class CContentShiftingEvasion(CFormatExploitEvasion):
	"""
	Content shifting attack
	"""
	def __init__(self,
			end2end_model : CClassifierEnd2EndMalware,
			preferable_extension_amount=0x200,
			iterations : int =100,
			is_debug: bool = False,
			random_init: bool = False,
			threshold: float = 0.5,
			penalty_regularizer: float = 0,
			chunk_hyper_parameter : int = None,
	):
		"""
		Create the content shifting attack

		Parameters
		----------
		end2end_model : CClassifierEnd2EndMalware
			the target end-to-end model
		preferable_extension_amount: int, optional, default 512
			how many bytes to inject, modulo the file alignment
		iterations : int, optional, default 100
			the number of iterations of the optimizer
		is_debug : bool, optional, default False
			if True, prints debug information during the optimization
		random_init : bool, optional, default False
			if True, it randomizes the locations set by index_to_perturb before starting the optimization
		threshold : float, optional, default 0
			the detection threshold to bypass. Default is 0
		penalty_regularizer : float
			the reularization parameter, Default is 0
		chunk_hyper_parameter : int, optional, default 256
			how many bytes to optimize at each round. Default is 256
		"""
		super(CContentShiftingEvasion, self).__init__(
			end2end_model=end2end_model,
			pe_header_extension=0,
			preferable_extension_amount=preferable_extension_amount,
			iterations=iterations,
			is_debug=is_debug,
			random_init=random_init,
			threshold=threshold,
			penalty_regularizer=penalty_regularizer,
			chunk_hyper_parameter=chunk_hyper_parameter
		)


import os
from os import listdir

import magic
import numpy as np
from secml.array import CArray
from secml.testing import CUnitTest

from secml_malware.models import CClassifierEmber
from secml_malware.models.c_classifier_end2end_malware import CClassifierEnd2EndMalware
from secml_malware.models.malconv import MalConv


class End2EndBaseTests(CUnitTest):
	def setUp(self):

		self.max_length = 2 ** 20
		self.padding_value = 256
		self.root_module_path = os.path.dirname(
			os.path.dirname(os.path.dirname(__file__))
		)
		self.classifier = CClassifierEnd2EndMalware(MalConv())
		self.malconv_plus = CClassifierEnd2EndMalware(MalConv(), plus_version=True)
		self.surrogate_classifier = CClassifierEnd2EndMalware(MalConv())
		self.ember_path = os.path.join(
			self.root_module_path, "../data/trained/pretrained_malconv.pth"
		)
		self.surrogate_path = os.path.join(
			self.root_module_path, "../data/trained/pretrained_malconv.pth"
		)
		self.malware_folder = os.path.join(
			self.root_module_path, "../data/malware_samples/test_folder"
		)
		self.goodware_folder = os.path.join(
			self.root_module_path, "../data/goodware_samples/"
		)
		self.single_malware_path = os.path.join(
			self.root_module_path, "../data/malware_samples/test_malware"
		)
		self.baseline = np.array([np.zeros(self.max_length) + self.padding_value])
		malware_file_paths = [os.path.join(self.malware_folder, f) for f in listdir(self.malware_folder)]
		self.complete_malware_paths = []
		X = []
		y = []
		for complete_path in malware_file_paths:
			if not os.path.isfile(complete_path):
				continue
			if "PE32" not in magic.from_file(complete_path): continue
			with open(complete_path, "rb") as malware:
				print(f'>Using {os.path.dirname(complete_path)}')
				self.complete_malware_paths.append(complete_path)
				code = MalConv.bytes_to_numpy(malware.read(), self.max_length, 256, False)
			X.append(code)
			y.append(1)
		# X.append(self.baseline[0])
		# y.append(0)
		self.X = CArray(X)
		self.Y = y
		with open(self.single_malware_path, "rb") as f:
			self.byte_malware = bytearray(f.read())
		self.malware = np.array(
			[MalConv.bytes_to_numpy(self.byte_malware, self.max_length, 256, False)]
		)


class EmberBaseTests(CUnitTest):
	def setUp(self):

		self.root_module_path = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))

		self.ember_path = os.path.join(self.root_module_path, "../data/trained/ember_model.txt")
		self.classifier = CClassifierEmber(tree_path=self.ember_path)

		self.malware_folder = os.path.join(self.root_module_path, "../data/malware_samples/test_folder")
		self.goodware_folder = os.path.join(self.root_module_path, "../data/goodware_samples/")

		X = []
		y = []
		for f in listdir(self.malware_folder):
			complete_path = os.path.join(self.malware_folder, f)
			if not os.path.isfile(complete_path):
				continue
			if "PE32" not in magic.from_file(complete_path): continue
			max_length = 0
			with open(complete_path, "rb") as malware:
				print(f'>Using {f}')
				# code = MalConv.bytes_to_numpy(malware.read(), self.max_length, 256, False)
				code = malware.read()
				max_length = max(max_length, len(code))
			X.append(code)
			y.append(1)

		self.X = [MalConv.bytes_to_numpy(x, max_length, 256, False) for x in X]
		self.X = CArray(self.X)
		self.Y = y



import lief
import numpy as np

from secml_malware.attack.whitebox.c_fast_gradient_sign_evasion import CFastGradientSignMethodEvasion
from secml_malware.models import CClassifierEnd2EndMalware
from secml_malware.utils.pe_operations import create_int_list_from_x_adv


class CKreukEvasion(CFastGradientSignMethodEvasion):
	def __init__(
			self,
			end2end_model: CClassifierEnd2EndMalware,
			how_many_padding_bytes: int,
			epsilon: float,
			iterations: int = 100,
			is_debug: bool = False,
			threshold: float = 0.5,
			p_norm: float = np.infty,
			compute_slack: bool = True,
			store_checkpoints : int = None
	):
		"""
		Create the padding attack by Kreuk et al. https://arxiv.org/abs/1802.04528

		Parameters
		----------
		end2end_model : CClassifierEnd2EndMalware
			the target end-to-end model
		how_many_padding_bytes: int, optional, default 512
			how many padding bytes
		epsilon : float
			the distortion amount
		iterations : int, optional, default 100
			the number of iterations of the optimizer
		is_debug : bool, optional, default False
			if True, prints debug information during the optimization
		threshold : float, optional, default 0
			the detection threshold to bypass. Default is 0
		p_norm : float, optional, default np.infty:
			the norm to use for compute the attack
		compute_slack : bool, optional, default True
			if True, uses also the slack space between sections. Default True
		store_checkpoints: int, optional, default None
			if set, it reconstruct the samples after the number of iteration specified. Default None
		"""
		super(CKreukEvasion, self).__init__(
			end2end_model=end2end_model,
			indexes_to_perturb=[],
			epsilon=epsilon,
			iterations=iterations,
			is_debug=is_debug,
			threshold=threshold,
			penalty_regularizer=0,
			p_norm=p_norm,
			store_checkpoints=store_checkpoints
		)
		self.how_many_padding_bytes = how_many_padding_bytes
		self.compute_slack = compute_slack

	def _run(self, x0, y0, x_init=None):
		padding = self._create_pading_indexes(x0)
		self.indexes_to_perturb = self._create_slack_indexes(x0) + padding if self.compute_slack else padding
		return super(CFastGradientSignMethodEvasion, self)._run(x0, y0, x_init=x_init)

	def _create_slack_indexes(self, x0):
		x_bytes = create_int_list_from_x_adv(x0, self.classifier.get_embedding_value(),
											 self.classifier.get_is_shifting_values())
		try:
			liefpe = lief.PE.parse(x_bytes)
		except:
			return []
		window_input_length = self.classifier.get_input_max_length()
		all_slack_space = []
		for s in liefpe.sections:
			if s.size > s.virtual_size:
				all_slack_space.extend(list(range(min(window_input_length, s.offset + s.virtual_size),
												  min(window_input_length, s.offset + s.size))))
		return all_slack_space

	def _create_pading_indexes(self, x0):
		invalid_value = 256 if self.invalid_pos == -1 else self.invalid_pos
		padding_positions = x0.find(x0 == invalid_value)
		if not padding_positions:
			return []
		else:
			return list(
				range(
					padding_positions[0],
					min(x0.size, padding_positions[0] + self.how_many_padding_bytes),
				)
			)



import struct
import numpy as np
import torch
from secml.settings import SECML_PYTORCH_USE_CUDA

from secml_malware.attack.whitebox.c_discretized_bytes_evasion import CDiscreteBytesEvasion
from secml_malware.models import CClassifierEnd2EndMalware

use_cuda = torch.cuda.is_available() and SECML_PYTORCH_USE_CUDA


class CHeaderEvasion(CDiscreteBytesEvasion):
	"""Creates the attack that perturbs the header of a Windows PE malware.
	"""

	def __init__(
			self,
			end2end_model: CClassifierEnd2EndMalware,
			index_to_perturb: list = None,
			iterations: int = 100,
			is_debug: bool = False,
			random_init: bool = False,
			optimize_all_dos: bool = False,
			threshold: float = 0,
			penalty_regularizer: int = 0
	):
		"""
		Creates the evasion object

		Parameters
		----------
		end2end_model : CClassifierEnd2EndMalware
			the target end-to-end model
		index_to_perturb : list
			a list containing the index to perturb inside the samples
		iterations : int, optional, default 100
			the number of iterations of the optimizer
		is_debug : bool, optional, default False
			if True, prints debug information during the optimization
		random_init : bool, optional, default False
			if True, it randomizes the locations set by index_to_perturb before starting the optimization
		optimize_all_dos : bool, optional, default False
			if True, set as editable all the DOS header, not only the specified portion
		threshold : float, optional, default 0
			the detection threshold to bypass. Default is 0
		penalty_regularizer : float
			the reularization parameter, Default is 0
		"""

		if index_to_perturb is None:
			index_to_perturb = [i for i in range(2, 0x3C)]
		super(CHeaderEvasion, self).__init__(
			end2end_model,
			index_to_perturb,
			iterations,
			is_debug,
			random_init,
			threshold,
			penalty_regularizer
		)
		self.optimize_all_dos = optimize_all_dos

	def _set_dos_indexes(self, x_init):
		if self.optimize_all_dos:
			pe_position = x_init[0x3C:0x40].tondarray().astype(np.uint16)[0]
			if self.shift_values:
				pe_position = np.array([p - 1 for p in pe_position])
			pe_position = struct.unpack("<I", bytes(pe_position.astype(np.uint8)))[0]
			self.indexes_to_perturb = [i for i in range(2, 0x3C)] + [
				i for i in range(0x40, pe_position)
			]
			self._how_many = len(self.indexes_to_perturb)
			if self.is_debug:
				print(f"PE POSITION: {pe_position}, perturbing {self._how_many}")

	def _run(self, x0, y0, x_init=None):
		self._set_dos_indexes(x_init)
		return super(CHeaderEvasion, self)._run(x0, y0, x_init)



import copy
import os
import pickle
import lief
import magic
import numpy as np
from secml.array import CArray
from secml_malware.attack.blackbox.c_blackbox_problem import CBlackBoxProblem
from secml_malware.attack.blackbox.c_wrapper_phi import CWrapperPhi


class CGammaEvasionProblem(CBlackBoxProblem):
	"""
	GAMMA padding attack
	"""

	def __init__(self,
				 section_population: list,
				 model_wrapper: CWrapperPhi,
				 population_size: int,
				 penalty_regularizer: float,
				 iterations: int,
				 seed: int = None,
				 is_debug: bool = False,
				 hard_label: bool = False,
				 threshold: float = 0.5,
				 loss: str = 'l1'):
		"""
		Creates the GAMMA padding attack.

		Parameters
		----------
		section_population : list
			a list containing all the goodware sections to inject
		model_wrapper : CWrapperPhi
			the target model, wrapped inside a CWrapperPhi
		population_size : int
			the population size generated at each round by the genetic algorithm
		penalty_regularizer: float
			the regularization parameter used for the size constraint
		iterations : int, optional, default 100
			the total number of iterations, default 100
		seed : int, optional, default None
			specifies an initialization seed for the random. None for not using determinism
		is_debug : bool, optional, default False
			if True, it prints messages while optimizing. Default is False
		hard_label : bool, optional default False
			if True, the problem will use only binary labels instead. Infinity will be used for non-evasive samples.
		threshold : float, optional, default 0
			the detection threshold. Leave 0 to test the degradation of the model until the end of the algorithm.
		loss : str, optional, default l1
			The loss function used as objective function
		"""
		super(CGammaEvasionProblem, self).__init__(model_wrapper,
												   len(section_population),
												   population_size,
												   penalty_regularizer,
												   iterations,
												   seed,
												   is_debug,
												   hard_label,
												   threshold,
												   loss)

		self.section_population = section_population
		self.payload_max_size = sum([len(s) for s in section_population])

	def apply_feasible_manipulations(self, t: np.ndarray, x: CArray) -> CArray:
		"""
		Applies the padding manipulation.

		Parameters
		----------
		t : np.ndarray
			the vector of parameters specifying how much content must be included
		x : CArray
			the original malware

		Returns
		-------
		CArray
			the adversarial malware
		"""
		x_adv = copy.deepcopy(x)
		for i in range(t.shape[-1]):
			content = self.section_population[i]
			content_to_append = content[:int(round(len(content) * t[i]))]
			x_adv = x_adv.append(content_to_append)
		x_adv = x_adv.reshape((1, x_adv.shape[-1]))
		return x_adv

	@classmethod
	def create_section_population_from_list(cls, folder: str, what_from_who: list) -> list:
		"""
		Create the section population from files contained in a specified folder

		Parameters
		----------
		folder : str
			the folder containing the file to open
		what_from_who : list
			a list of file (section name, file names) that specifies what extract from who
		Returns
		-------
		list
			the section population list
		"""
		section_population = []
		for entry in what_from_who:
			what, who = entry
			path = os.path.join(folder, who)
			lief_pe_file = lief.PE.parse(path)
			for s in lief_pe_file.sections:
				if s.name == what:
					section_population.append(s.content)
		return section_population

	@classmethod
	def create_section_population_from_folder(cls, folder: str, how_many: int, sections_to_extract: list = None,
											  cache_file: str = None,
											  size_lower_bound: int = None) -> (list, list):
		"""
		Extract sections from a given folder

		Parameters
		----------
		folder : str
			the folder containing programs used for extracting sections
		how_many : int
			how many sections to extract in general
		sections_to_extract : list, optional, default None
			the list of section names to use. If None, it will extract only .data sections
		cache_file : str, optional, default None
			if set, it stores which section from what program has been used inside a pickled object, stored in path
		size_lower_bound : int, optional, default None
			if set, it will discard all the sections whose content length is less that such parameter

		Returns
		-------
		list, list
			the section population and what has been extracted from who
		"""
		if sections_to_extract is None:
			sections_to_extract = [".data"]
		section_population = []
		counter = 0
		what_from_who = []
		if cache_file and os.path.isfile(cache_file):
			with open(cache_file, 'rb') as section_file:
				file_to_consider = pickle.load(section_file)
				file_to_consider = [f[1] for f in file_to_consider]
		else:
			file_to_consider = os.listdir(folder)
		for filename in file_to_consider:
			path = os.path.join(folder, filename)
			if "PE" not in magic.from_file(path):
				continue
			lief_pe_file = lief.PE.parse(path)
			for s in lief_pe_file.sections:
				if s.name in sections_to_extract:
					if size_lower_bound and len(s.content) < size_lower_bound:
						continue
					if len(s.content) == 0:
						continue
					section_population.append(s.content)
					what_from_who.append((s.name, filename))
					counter += 1
			if counter >= how_many:
				break
		section_population = section_population[:how_many]
		if cache_file and not os.path.isfile(cache_file):
			with open(cache_file, 'wb') as section_file:
				pickle.dump(what_from_who, section_file)
		return section_population, what_from_who



from abc import abstractmethod

from secml.array import CArray
from secml.ml.classifiers import CClassifier

from secml_malware.models import CClassifierEnd2EndMalware, CClassifierEmber
from secml_malware.models.c_classifier_sorel_net import CClassifierSorel


class CWrapperPhi:
	"""
	Abstract class that encapsulates a model for being used in a black-box way.
	"""

	def __init__(self, model: CClassifier):
		"""
		Creates the wrapper.

		Parameters
		----------
		model : CClassifier
		The model to wrap
		"""
		self.classifier = model

	@abstractmethod
	def extract_features(self, x: CArray):
		"""
		It maps the input sample inside the feature space of the wrapped model.

		Parameters
		----------
		x : CArray
			The sample in the input space.
		Returns
		-------
		CArray
			The feature space representation of the input sample.
		"""
		raise NotImplementedError("This method is abstract, you should implement it somewhere else!")

	def predict(self, x: CArray, return_decision_function: bool = True):
		"""
		Returns the prediction of the sample (in input space).

		Parameters
		----------
		x : CArray
			The input sample in input space.
		return_decision_function : bool, default True
			If True, it also returns the decision function value, rather than only the label.
			Default is True.
		Returns
		-------
		CArray, (CArray)
			Returns the label of the sample.
			If return_decision_function is True, it also returns the output of the decision function.
		"""
		x = x.atleast_2d()
		# feature_vectors = []
		# for i in range(x.shape[0]):
		# 	x_i = x[i, :]
		# 	padding_position = x_i.find(x_i == 256)
		# 	if padding_position:
		# 		x_i = x_i[0, :padding_position[0]]
		# 	feature_vectors.append(self.extract_features(x_i))
		# feature_vectors = CArray(feature_vectors)
		feature_vectors = self.extract_features(x)
		return self.classifier.predict(feature_vectors, return_decision_function=return_decision_function)


class CEmberWrapperPhi(CWrapperPhi):
	"""
	Class that wraps a GBDT classifier with EMBER feature set.
	"""

	def __init__(self, model: CClassifierEmber):
		"""
		Creates the wrapper of a CClassifierEmber.

		Parameters
		----------
		model : CClassifierEmber
		The GBDT model to wrap
		"""
		if not isinstance(model, CClassifierEmber):
			raise ValueError(f"Input model is {type(model)} and not CClassifierEmber")
		super().__init__(model)

	def extract_features(self, x):
		"""
		It extracts the EMBER hand-crafted features

		Parameters
		----------
		x : CArray
			The sample in the input space.
		Returns
		-------
		CArray
			The feature space representation of the input sample.
		"""
		x = x.atleast_2d()
		clf: CClassifierEmber = self.classifier
		feature_vectors = CArray.zeros((x.shape[0], 2381))
		for i in range(x.shape[0]):
			x_i = x[i, :]
			padding_positions = x_i.find(x_i == 256)
			if padding_positions:
				feature_vectors[i, :] = clf.extract_features(x_i[0, :padding_positions[0]])
			else:
				feature_vectors[i, :] = clf.extract_features(x_i)
		return feature_vectors


class CSorelWrapperPhi(CWrapperPhi):
	"""
	Class that wraps a deep neural network with SOREL feature set.
	"""

	def __init__(self, model: CClassifierSorel, use_only_malware_score=True):
		"""
		Creates the wrapper of a CClassifierSorel.

		Parameters
		----------
		model : CClassifierSorel
		The deep neural network model to wrap
		"""
		if not isinstance(model, CClassifierSorel):
			raise ValueError(f"Input model is {type(model)} and not CClassifierEmber")
		super().__init__(model)
		self._malware_score = use_only_malware_score

	def extract_features(self, x):
		"""
		It extracts the EMBER hand-crafted features

		Parameters
		----------
		x : CArray
			The sample in the input space.
		Returns
		-------
		CArray
			The feature space representation of the input sample.
		"""
		x = x.atleast_2d()
		clf: CClassifierSorel = self.classifier
		feature_vectors = CArray.zeros((x.shape[0], 2381))
		for i in range(x.shape[0]):
			x_i = x[i, :]
			padding_positions = x_i.find(x_i == 256)
			if padding_positions:
				feature_vectors[i, :] = clf.extract_features(x_i[0, :padding_positions[0]])
			else:
				feature_vectors[i, :] = clf.extract_features(x_i)
		return feature_vectors

	def predict(self, x: CArray, return_decision_function: bool = True):
		result = super(CSorelWrapperPhi, self).predict(x, return_decision_function=return_decision_function)
		if self._malware_score:
			return result[:, 0, 1]
		return result


class CEnd2EndWrapperPhi(CWrapperPhi):
	"""
	Class that wraps an end-to-end model
	"""

	def __init__(self, model: CClassifierEnd2EndMalware):
		"""
		Creates the wrapper of a CClassifierEnd2EndMalware.

		Parameters
		----------
		model : CClassifierEnd2EndMalware
		The end to end model to wrap
		"""
		if not isinstance(model, CClassifierEnd2EndMalware):
			raise ValueError(f"Input model is {type(model)} and not CClassifierEnd2EndMalware")
		super().__init__(model)

	def extract_features(self, x):
		"""
		Crops and pads the input sample for being passed to the network.

		Parameters
		----------
		x : CArray
			The sample in the input space.
		Returns
		-------
		CArray
			The feature space representation of the input sample.
		"""
		clf: CClassifierEnd2EndMalware = self.classifier
		x = x.atleast_2d()
		padded_x = CArray.zeros((x.shape[0], clf.get_input_max_length())) + clf.get_embedding_value()
		for i in range(x.shape[0]):
			x_i = x[i, :]
			length = min(x_i.shape[-1], clf.get_input_max_length())
			padded_x[i, :length] = x_i[0, :length] + clf.get_is_shifting_values()
		return padded_x



import struct
import string
import seaborn
import numpy as np
import matplotlib.pyplot as plt

seaborn.set()
COFF_LENGTH = 24
SECT_ENTRY_LEN = 40


def _pick_color(max_v, value):
	color = str.format("{:02x}", 255 - int(value / max_v * 255))
	rgb = "#ff{}ff".format(color)
	return rgb


def plot_bins(itgs: np.ndarray, n_bins: int = 256, cut_index: int = None, force_plot: bool = True):
	"""Plot contribution in histogram
	
	Parameters
	----------
	itgs : numpy array
		The array containing the results of Integrated Gradients.
	n_bins : int, optional, default 256
		How many bins fot the histogram, default 256
	cut_index : int, optional, default None
		limit how many location should be used, default None
	force_plot : bool, optional, default True
		Should show plot on screen? Default True
	"""
	attr = itgs[1, :].tolist()[0]
	n_bytes = len(attr) // n_bins
	positives = [i if i >= 0 else 0 for i in attr]
	negatives = [i if i < 0 else 0 for i in attr]
	x = range(n_bins)
	if cut_index:
		positives = positives[:cut_index]
		negatives = negatives[:cut_index]
		n_bins = cut_index // n_bytes
		x = x[:n_bins]
	y_p = [np.sum(positives[i * n_bytes: (i + 1) * n_bytes]) for i in range(n_bins)]
	y_n = [np.sum(negatives[i * n_bytes: (i + 1) * n_bytes]) for i in range(n_bins)]
	plt.bar(x, y_p, color="b")
	plt.bar(x, y_n, color="r")
	plt.xlabel("Segment of {} bytes".format(n_bytes))
	plt.ylabel("Contribution")
	plt.title("Sum of contributions, divided in {} bins".format(n_bins))
	if force_plot:
		plt.show()


def plot_range_histogram(itgs: np.ndarray, start_idx: int = 0, how_many: int = None, force_plot: bool = True):
	"""Plot chunk of bytes with their relevance, according to the integrated gradient technique

	Parameters
	----------

	itgs : numpy array
		array containing the result of Integrated gradient
	start_idx : int
		starting index for visualizing the results, default 0
	how_many : int
		how many bytes represent inside the plot (default : None, means use all)
	force_plot : bool, optional, default True
		Should show plot on screen? Default True
	"""
	attr = itgs[1, :].tolist()[0]
	if start_idx > len(attr):
		raise ValueError()
	stop_index = len(attr) if how_many is None else how_many + start_idx
	if stop_index < start_idx:
		raise ValueError()
	if stop_index > len(attr):
		raise ValueError()
	attr = attr[start_idx:stop_index]
	positives = [i if i >= 0 else 0 for i in attr]
	negatives = [i if i < 0 else 0 for i in attr]
	x = range(how_many)
	plt.bar(x, positives, color="r")
	plt.bar(x, negatives, color="b")
	plt.title("Offset histogram, from {} to {} ".format(start_idx, stop_index))
	if force_plot:
		plt.show()


def plot_header_contribution_histogram(bytestring_program: bytearray, itgs: np.ndarray, percentage: bool = True,
									   force_plot: bool = True):
	"""Plot integrated gradient results, divided by sections

	Parameters
	----------
	bytestring_program: bytearray
		the program as bytearray
	itgs : numpy array
		array containing the result of Integrated gradient
	percentage :bool
		display percentage instead of absolute values (default: True)
	force_plot : bool
		Should show the results? (default:True)
	"""
	pe_position = struct.unpack("<I", bytestring_program[0x3C: 0x3C + 4])[0]
	n_sects = struct.unpack(
		"<I", bytestring_program[pe_position + 6: pe_position + 8] + b"\x00\x00"
	)[0]
	opt_h_len = struct.unpack(
		"<I", bytestring_program[pe_position + 20: pe_position + 22] + b"\x00\x00"
	)[0]
	sect_offset = opt_h_len + pe_position + COFF_LENGTH
	sect_end = sect_offset + SECT_ENTRY_LEN * n_sects
	mean_dos = np.mean(itgs[0, 0:pe_position].tondarray())
	mean_header_coff = np.sum(itgs[0, pe_position: pe_position + 24].tondarray())
	mean_header_optional = np.sum(
		itgs[0, pe_position + COFF_LENGTH: pe_position + COFF_LENGTH + opt_h_len].tondarray()
	)
	mean_section_table = np.sum(itgs[0, sect_offset:sect_end].tondarray())
	names = ["DOS Header", "COFF Header", "Optional Header", "Section Headers"]
	to_plot = [mean_dos, mean_header_coff, mean_header_optional, mean_section_table]
	sect_name_length = 8
	for i in range(n_sects):
		offset_index = sect_offset + i * SECT_ENTRY_LEN + sect_name_length + 12
		size_index = sect_offset + i * SECT_ENTRY_LEN + sect_name_length + 8
		offset = struct.unpack("<I", bytestring_program[offset_index: offset_index + 4])[0]
		size = struct.unpack("<I", bytestring_program[size_index: size_index + 4])[0]
		name = str(
			bytestring_program[
			sect_offset + i * SECT_ENTRY_LEN: sect_offset + i * SECT_ENTRY_LEN + 8
			]
				.decode("utf-8")
				.rstrip("\x00")
		)
		mean = np.sum(itgs[0, offset: offset + size].tondarray())
		to_plot.append(mean)
		names.append(name)
	if percentage:
		to_plot = to_plot / np.linalg.norm(to_plot, ord=2)
	plt.figure()
	x = range(len(names))
	positives = [i if i >= 0 else 0 for i in to_plot]
	negatives = [i if i < 0 else 0 for i in to_plot]
	plt.bar(x, positives, width=0.2, color="r")
	plt.bar(x, negatives, width=0.2, color="b")
	plt.yticks(fontsize=22)
	plt.xticks(x, names, fontsize=22, rotation=45, ha="right")
	xs = np.linspace(-1, 7, 4)
	plt.plot(xs, np.zeros(len(xs)), "k")
	ax = plt.gca()
	ax.set_facecolor((1.0, 1.0, 1.0))
	plt.title(
		"Sum of each contribution,divided into headers and sections\n", fontsize=25
	)
	if force_plot:
		plt.show()


def plot_code_segment(
		pe_file: list,
		start: int,
		stop: int,
		itgs: np.ndarray,
		title: str,
		show_positives: bool = True,
		show_negatives: bool = False,
		force_plot: bool = True,
		width: int = 16,
		percentage: bool = True,
):
	"""Plot contribution of chunks of bytes.

	Parameters
	----------
	pe_file : list
		list of bytes
	start : int
		starting index for segment to plot
	stop : int
		stop index for segment to plot
	itgs : numpy array
		array containing result of integrated gradients
	title : str
		plot title
	show_positives : bool
		show positives contributes (default:True)
	show_negatives : bool
		show negative contributes (default:False}
	force_plot : bool
		show plot (default:True)
	width : int
		how many byte per row of the heatmap (default:16)
	percentage : bool
		display percentage instead of absolute values (default: True)
	"""
	grad_section = itgs[1, start:stop].tolist()[0]
	text = [
		hex(i) if chr(i) not in string.ascii_letters else chr(i)
		for i in pe_file[start:stop]
	]
	cols = width
	tot_len = len(text)
	row = tot_len // cols
	if tot_len % cols:
		rem = tot_len % cols
		text.extend(["" for _ in range(cols - rem)])
		grad_section.extend([0 for _ in range(cols - rem)])
		row = row + 1
	text = np.array(text)
	grad_section = np.array(grad_section) / (
		np.linalg.norm(grad_section) if percentage else 1
	)

	if not show_positives:
		grad_section[grad_section > 0] = 0
	if not show_negatives:
		grad_section[grad_section < 0] = 0

	grad_section = grad_section.reshape((row, cols))
	text = text.reshape((row, cols))
	plt.figure()
	hmap = seaborn.heatmap(
		grad_section,
		annot=text,
		fmt="",
		cmap="seismic",
		center=0,
		annot_kws={"size": 18},
	)
	hmap.collections[0].colorbar.ax.tick_params(labelsize=20)
	hmap.set_title(title, fontsize=25)
	if force_plot:
		plt.show()



import torch
import numpy as np
from secml.array import CArray
from secml.settings import SECML_PYTORCH_USE_CUDA
from secml_malware.attack.whitebox import CEnd2EndMalwareEvasion
from secml_malware.models import CClassifierEnd2EndMalware

use_cuda = torch.cuda.is_available() and SECML_PYTORCH_USE_CUDA

class CFastGradientSignMethodEvasion(CEnd2EndMalwareEvasion):
	"""Creates the basic attack that implements the Fast Gradient Sign Method for the Windows malware domain.
	The original attack has been proposed by Goodfellow et al. (https://arxiv.org/abs/1412.6572)
	"""
	def __init__(
		self,
		end2end_model: CClassifierEnd2EndMalware,
		indexes_to_perturb: list,
		epsilon: float,
		iterations: int = 100,
		is_debug: bool = False,
		random_init: bool = False,
		threshold : float = 0.5,
		penalty_regularizer : float = 0,
		p_norm : float = np.infty,
		store_checkpoints : int = None
	):
		super(CFastGradientSignMethodEvasion, self).__init__(
			end2end_model=end2end_model,
			indexes_to_perturb=indexes_to_perturb,
			iterations=iterations,
			is_debug=is_debug,
			random_init=random_init,
			threshold=threshold,
			penalty_regularizer=penalty_regularizer,
			store_checkpoints = store_checkpoints
		)
		self.epsilon = epsilon
		self.p_norm = p_norm

	def compute_penalty_term(self, original_x: CArray, adv_x: CArray, par: float):
		penalty_term = torch.autograd.Variable(torch.tensor([0]))
		if use_cuda:
			penalty_term = penalty_term.cuda()
		return penalty_term

	def loss_function_gradient(self, original_x : CArray, adv_x : CArray, penalty_term : torch.Tensor):
		y = self.classifier.embedding_predict(adv_x)
		malware_class = torch.tensor([1.0])
		if use_cuda:
			malware_class = malware_class.cuda()
		loss = torch.nn.functional.binary_cross_entropy(y, malware_class)
		g = torch.autograd.grad(loss, adv_x)[0]
		g = torch.transpose(g, 1, 2)[0]
		return g

	def optimization_solver(self, E, gradient_f, index_to_consider, x_init):
		gradient_result = self._internal_fsgm_solver(gradient_f).transpose(0,1)
		x_init[0, :, index_to_consider] = x_init[0, :, index_to_consider] + gradient_result[:, index_to_consider]
		return x_init

	def _internal_fsgm_solver(self, gradient_f):
		g = gradient_f / torch.norm(gradient_f) if not torch.equal(torch.zeros(gradient_f.shape), gradient_f) else torch.zeros(gradient_f.shape)
		if self.p_norm == 2:
			return  self.epsilon * g
		elif self.p_norm == np.infty:
			return self.epsilon * torch.sign(g)
		raise NotImplementedError(f"{self.p_norm}-norm not yet implemented")

	def infer_step(self, x_init):
		confidence = self.classifier.embedding_predict(x_init)
		return confidence.item()

	def invert_feature_mapping(self, x, x_adv):
		E = self._get_embedded_byte_matrix()
		byte_malware = torch.zeros(x.shape[-1]) + self.classifier.get_embedding_value()
		E[self.invalid_pos] = torch.tensor([np.infty for _ in range(self.classifier.get_embedding_size())])
		reconstructed_x = x.tondarray()
		for i in self.indexes_to_perturb:
			x_i = x_adv[0, :, i]
			byte_to_consider = torch.tensor([torch.norm(e - x_i, p=2) for e in E]).argmin()
			byte_malware[i] = byte_to_consider
		reconstructed_x[0, self.indexes_to_perturb] = byte_malware[self.indexes_to_perturb]
		return CArray(reconstructed_x)

	def apply_feature_mapping(self, x : CArray):
		return self.classifier.embed(x.tondarray())



import struct

import torch
from secml.settings import SECML_PYTORCH_USE_CUDA

from secml_malware.attack.whitebox.c_discretized_bytes_evasion import CDiscreteBytesEvasion
from secml_malware.models import CClassifierEnd2EndMalware
from secml_malware.utils.pe_operations import create_indexes_for_header_fields_manipulations

use_cuda = torch.cuda.is_available() and SECML_PYTORCH_USE_CUDA


class CHeaderFieldsEvasion(CDiscreteBytesEvasion):
	"""
	Creates a black-box attack that perturbs 18 bytes inside the COFF and Optional Header of an executable.
	"""

	def __init__(
			self,
			end2end_model: CClassifierEnd2EndMalware,
			index_to_perturb: list = None,
			iterations: int = 100,
			is_debug: bool = False,
			random_init: bool = False,
			optimize_all_dos: bool = False,
			threshold: float = 0,
			penalty_regularizer: int = 0
	):
		"""
		Creates the evasion object

		Parameters
		----------
		end2end_model : CClassifierEnd2EndMalware
			the target end-to-end model
		iterations : int, optional, default 100
			the number of iterations of the optimizer
		is_debug : bool, optional, default False
			if True, prints debug information during the optimization
		random_init : bool, optional, default False
			if True, it randomizes the locations set by index_to_perturb before starting the optimization
		optimize_all_dos : bool, optional, default False
			if True, set as editable all the DOS header, not only the specified portion
		threshold : float, optional, default 0
			the detection threshold to bypass. Default is 0
		penalty_regularizer : float
			the regularization parameter. Default is 0
		"""

		super(CHeaderFieldsEvasion, self).__init__(
			end2end_model,
			index_to_perturb,
			iterations,
			is_debug,
			random_init,
			threshold,
			penalty_regularizer
		)
		self.optimize_all_dos = optimize_all_dos

	def _set_header_fields_indexes(self, x_init):
		pe_index = struct.unpack('<I', bytes(x_init[0, 60:64].astype(int).tolist()[0]))[0]
		self.indexes_to_perturb = create_indexes_for_header_fields_manipulations(pe_index)
		return self.indexes_to_perturb

	def _run(self, x0, y0, x_init=None):
		self._set_header_fields_indexes(x_init)
		return super(CHeaderFieldsEvasion, self)._run(x0, y0, x_init)



class FileNotExistsException(Exception):
	def __init__(self, message):
		super(FileNotExistsException, self).__init__(message)


class FolderNotExistsException(Exception):
	def __init__(self, message):
		super(FolderNotExistsException, self).__init__(message)

class OptimizerNotSetException(Exception):
	def __init__(self, message):
		super(OptimizerNotSetException, self).__init__(message)


# Copyright 2020, Sophos Limited. All rights reserved.
#
# 'Sophos' and 'Sophos Anti-Virus' are registered trademarks of
# Sophos Limited and Sophos Group. All other product and company
# names mentioned are trademarks or registered trademarks of their
# respective owners.
import os

import torch
from ember import PEFeatureExtractor
from secml.array import CArray
from secml.ml import CClassifier
from secml.settings import SECML_PYTORCH_USE_CUDA
from torch import nn
import numpy as np

use_cuda = torch.cuda.is_available() and SECML_PYTORCH_USE_CUDA


class SorelNet(nn.Module):
	"""
	This is a simple network loosely based on the one used in ALOHA: Auxiliary Loss Optimization for Hypothesis Augmentation (https://arxiv.org/abs/1903.05700)
	Note that it uses fewer (and smaller) layers, as well as a single layer for all tag predictions, performance will suffer accordingly.
	"""

	def __init__(self, use_malware=True, use_counts=True, use_tags=True, n_tags=None, feature_dimension=2381,
				 layer_sizes=None):
		self.use_malware = use_malware
		self.use_counts = use_counts
		self.use_tags = use_tags
		self.n_tags = n_tags
		if self.use_tags and self.n_tags == None:
			raise ValueError("n_tags was None but we're trying to predict tags. Please include n_tags")
		super().__init__()
		p = 0.05
		layers = []
		if layer_sizes is None: layer_sizes = [512, 512, 128]
		for i, ls in enumerate(layer_sizes):
			if i == 0:
				layers.append(nn.Linear(feature_dimension, ls))
			else:
				layers.append(nn.Linear(layer_sizes[i - 1], ls))
			layers.append(nn.LayerNorm(ls))
			layers.append(nn.ELU())
			layers.append(nn.Dropout(p))
		self.model_base = nn.Sequential(*tuple(layers))
		self.malware_head = nn.Sequential(nn.Linear(layer_sizes[-1], 1),
										  nn.Sigmoid())
		self.count_head = nn.Linear(layer_sizes[-1], 1)
		self.sigmoid = nn.Sigmoid()
		self.tag_head = nn.Sequential(nn.Linear(layer_sizes[-1], 64),
									  nn.ELU(),
									  nn.Linear(64, 64),
									  nn.ELU(),
									  nn.Linear(64, n_tags),
									  nn.Sigmoid())

	def forward(self, data):
		rv = {}
		base_result = self.model_base.forward(data)
		if self.use_malware:
			rv['malware'] = self.malware_head(base_result)
		if self.use_counts:
			rv['count'] = self.count_head(base_result)
		if self.use_tags:
			rv['tags'] = self.tag_head(base_result)
		return rv


class CClassifierSorel(CClassifier):
	def __init__(self, model_path, use_malware=True, use_counts=True, use_tags=True, n_tags=11,
				 feature_dimension=2381, layer_sizes=None):
		super().__init__()
		self._sorel = SorelNet(use_malware, use_counts, use_tags, n_tags, feature_dimension, layer_sizes)
		if not os.path.isfile(model_path):
			raise FileNotFoundError(f'{model_path} not exists')
		self.load_model(model_path)

	def extract_features(self, x: CArray):
		extractor = PEFeatureExtractor(2)
		x_bytes = bytes(x.astype(np.int).tolist()[0])
		features = CArray([np.array(extractor.feature_vector(x_bytes), dtype=np.float32)])
		return features

	def load_model(self, model_path):
		state_dict = torch.load(model_path) if use_cuda else torch.load(model_path, map_location='cpu')
		self._sorel.load_state_dict(state_dict)
		if use_cuda:
			self._sorel.cuda()
		self._sorel.eval()
		self._classes = 3 + self._sorel.n_tags
		self._n_features = 2381
		return self

	def _fit(self, dataset, **kwargs):
		raise NotImplementedError("Fit is not implemented.")

	def _forward(self, x: CArray):
		x = torch.tensor(x.tondarray())
		if use_cuda:
			x = x.cuda()
		rv = self._sorel.forward(x)
		malware_score = rv['malware'].item() if 'malware' in rv else 0
		count_score = rv['count'].item() if 'count' in rv else 0
		tags_score = [r for r in rv['tags'].flatten().tolist()] if 'tags' in rv else [0] * self._sorel.n_tags
		confidence = CArray([1 - malware_score, malware_score, count_score] + tags_score)
		return confidence



import numpy
from secml.array import CArray

from secml_malware.attack.whitebox.c_discretized_bytes_evasion import CDiscreteBytesEvasion
from secml_malware.models import CClassifierEnd2EndMalware, End2EndModel
from secml_malware.utils.pe_operations import shift_section_by, shift_pe_header_by, create_int_list_from_x_adv


class CFormatExploitEvasion(CDiscreteBytesEvasion):

	def __init__(self,
				 end2end_model: CClassifierEnd2EndMalware,
				 preferable_extension_amount: int = 0x200,
				 pe_header_extension: int = 0x200,
				 iterations: int = 100,
				 is_debug: bool = False,
				 random_init: bool = False,
				 threshold: float = 0.5,
				 penalty_regularizer: float = 0,
				 chunk_hyper_parameter: int = None,
				 ):
		"""
		Creates the evasion object

		Parameters
		----------
		end2end_model : CClassifierEnd2EndMalware
			the target end-to-end model
		preferable_extension_amount : int, optional, default 512
			the number of bytes to inject before the first section, modulo file alignment. Default 512.
		pe_header_extension : int, optional, default 512
			the number of bytes to inject as new DOS header, modulo file alignment. Default 512.
		iterations : int, optional, default 100
			the number of iterations of the optimizer
		is_debug : bool, optional, default False
			if True, prints debug information during the optimization
		random_init : bool, optional, default False
			if True, it randomizes the locations set by index_to_perturb before starting the optimization
		threshold : float, optional, default 0
			the detection threshold to bypass. Default is 0
		penalty_regularizer : float
			the reularization parameter, Default is 0
		chunk_hyper_parameter : int, optional, default 256
			how many bytes to optimize at each round. Default is 256
		"""
		super(CFormatExploitEvasion, self).__init__(
			end2end_model=end2end_model,
			index_to_perturb=[],
			iterations=iterations,
			is_debug=is_debug,
			random_init=random_init,
			threshold=threshold,
			penalty_regularizer=penalty_regularizer,
			chunk_hyper_parameter=chunk_hyper_parameter
		)
		self.preferable_extension_amount = preferable_extension_amount
		self.pe_header_extension = pe_header_extension

	def _run(self, x0, y0, x_init=None):
		x_init, _ = self._craft_perturbed_c_array(x0)
		return super(CFormatExploitEvasion, self)._run(x0, y0, x_init)

	def _craft_perturbed_c_array(self, x0: CArray):
		x_init, indexes_to_perturb = self._generate_list_adv_example(x0)
		self.indexes_to_perturb = indexes_to_perturb
		x_init = CArray(
			[
				End2EndModel.list_to_numpy(
					x_init,
					self.classifier.get_input_max_length(),
					self.classifier.get_embedding_value(),
					self.classifier.get_is_shifting_values())
			]
		)
		return x_init, indexes_to_perturb

	def _generate_list_adv_example(self, x0):
		x_init = create_int_list_from_x_adv(x0, self.classifier.get_embedding_value(),
											self.classifier.get_is_shifting_values())
		x_init, index_to_perturb_sections = shift_section_by(x_init,
															 preferable_extension_amount=self.preferable_extension_amount)
		x_init, index_to_perturb_pe = shift_pe_header_by(x_init, preferable_extension_amount=self.pe_header_extension)
		indexes_to_perturb = index_to_perturb_pe + [i + len(index_to_perturb_pe) for i in
													index_to_perturb_sections]
		return x_init, indexes_to_perturb

	def create_real_sample_from_adv(self, original_file_path: str, x_adv: CArray,
									new_file_path: str = None) -> bytearray:
		"""
		Create a real adversarial example

		Parameters
		----------
		original_file_path : str
			the original malware sample
		x_adv : CArray
			the perturbed malware sample, as created by the optimizer
		new_file_path : str, optional, default None
			the path where to save the adversarial malware. Leave None to not save the result to disk

		Returns
		-------
		bytearray
			the adversarial malware, as string of bytes
		"""
		with open(original_file_path, 'rb') as f:
			code = bytearray(f.read())
		original_x = CArray([numpy.frombuffer(code, dtype=numpy.uint8).astype(numpy.uint16)])
		if self.classifier.get_is_shifting_values():
			original_x += self.classifier.get_is_shifting_values()
		x_init, index_to_perturb = self._generate_list_adv_example(original_x)
		x_init = CArray([x_init]).astype(numpy.uint8)
		x_init[0, index_to_perturb] = x_adv[0, index_to_perturb] - self.classifier.get_is_shifting_values()
		x_real = x_init[0, :].tolist()[0]
		x_real_adv = b''.join([bytes([i]) for i in x_real])
		if new_file_path:
			with open(new_file_path, 'wb') as f:
				f.write(x_real_adv)
		return x_real_adv



import numpy as np
from secml.array import CArray

from secml_malware.attack.blackbox.c_blackbox_problem import CBlackBoxProblem
from secml_malware.attack.blackbox.c_wrapper_phi import CWrapperPhi
from secml_malware.utils.pe_operations import shift_section_by, shift_pe_header_by, create_int_list_from_x_adv


class CBlackBoxFormatExploitEvasionProblem(CBlackBoxProblem):
	def __init__(
			self,
			model_wrapper: CWrapperPhi,
			population_size: int,
			preferable_extension_amount: int = 0x200,
			pe_header_extension: int = 0x200,
			iterations: int = 100,
			is_debug: bool = False,
			penalty_regularizer: float = 0,
			invalid_value: int = 256
	):
		super(CBlackBoxFormatExploitEvasionProblem, self).__init__(model_wrapper,
																   latent_space_size=preferable_extension_amount + pe_header_extension,
																   iterations=iterations,
																   population_size=population_size, is_debug=is_debug,
																   penalty_regularizer=penalty_regularizer)
		self.preferable_extension_amount = preferable_extension_amount
		self.pe_header_extension = pe_header_extension
		self.invalid_value = invalid_value
		self.indexes_to_perturb = []

	def init_starting_point(self, x: CArray) -> CArray:
		"""
		Initialize the problem, by setting the starting point.

		Parameters
		----------
		x : CArray
			the initial point

		Returns
		-------
		CArray
			the initial point (padded accordingly to remove trailing invalid values)
		"""
		_, indexes_to_perturb = self._craft_perturbed_c_array(x)
		self.indexes_to_perturb = indexes_to_perturb
		self.latent_space_size = len(indexes_to_perturb)
		return super(CBlackBoxFormatExploitEvasionProblem, self).init_starting_point(x)

	def _craft_perturbed_c_array(self, x0):
		x_init = create_int_list_from_x_adv(x0, self.invalid_value, False)
		x_init, index_to_perturb_sections = shift_section_by(x_init,
															 preferable_extension_amount=self.preferable_extension_amount)
		x_init, index_to_perturb_pe = shift_pe_header_by(x_init, preferable_extension_amount=self.pe_header_extension)
		indexes_to_perturb = index_to_perturb_pe + [i + len(index_to_perturb_pe) for i in
													index_to_perturb_sections]
		return x_init, indexes_to_perturb

	def apply_feasible_manipulations(self, t, x: CArray) -> CArray:
		"""
		Apply the format exploit practical manipulation on the input sample

		Parameters
		----------
		t : CArray
			the vector of manipulations in [0,1]
		x : CArray
			the input space sample to perturb

		Returns
		-------
		CArray:
			the adversarial malware
		"""
		byte_values = (t * 255).astype(np.int)
		x_adv, _ = self._craft_perturbed_c_array(x)
		for i, index in enumerate(self.indexes_to_perturb):
			x_adv[index] = byte_values[i]
		x_adv = CArray(x_adv)
		x_adv = x_adv.reshape((1, x_adv.shape[-1]))
		return CArray(x_adv)


class CBlackBoxContentShiftingEvasionProblem(CBlackBoxFormatExploitEvasionProblem):
	def __init__(
			self,
			model_wrapper: CWrapperPhi,
			population_size: int,
			bytes_to_inject: int = 0x200,
			iterations: int = 100,
			is_debug: bool = False,
			penalty_regularizer: float = 0,
			invalid_value: int = 256
	):
		super(CBlackBoxContentShiftingEvasionProblem, self).__init__(model_wrapper, population_size, bytes_to_inject, 0,
																	 iterations, is_debug, penalty_regularizer,
																	 invalid_value)


class CBlackBoxContentDOSExtensionProblem(CBlackBoxFormatExploitEvasionProblem):
	def __init__(
			self,
			model_wrapper: CWrapperPhi,
			population_size: int,
			bytes_to_inject: int = 0x200,
			iterations: int = 100,
			is_debug: bool = False,
			penalty_regularizer: float = 0,
			invalid_value: int = 256
	):
		super(CBlackBoxContentDOSExtensionProblem, self).__init__(model_wrapper, population_size, 0, bytes_to_inject,
																  iterations, is_debug, penalty_regularizer,
																  invalid_value)



import numpy as np
from secml.array import CArray

from secml_malware.attack.blackbox.c_blackbox_problem import CBlackBoxProblem
from secml_malware.attack.blackbox.c_wrapper_phi import CWrapperPhi


class CBlackBoxPaddingEvasionProblem(CBlackBoxProblem):
	"""
	Padding black-box problem
	"""
	def __init__(
			self,
			model_wrapper: CWrapperPhi,
			population_size: int,
			how_many_padding_bytes: int,
			iterations: int = 100,
			is_debug: bool = False,
			penalty_regularizer: float = 0,
			invalid_value: int = 256
	):
		"""
		Creates the Padding attack problem for the black-box engine.

		Parameters
		----------
		model_wrapper : CWrapperPhi
			the target model, wrapped inside a CWrapperPhi
		population_size : int
			the population size generated at each round by the genetic algorithm
		how_many_padding_bytes : int
			the number of padding bytes to append
		iterations : int, optional, default 100
			the total number of iterations, default 100
		is_debug : bool, optional, default False
			if True, it prints messages while optimizing. Default is False
		penalty_regularizer : float, optional, default 0
			the penalty regularizer for the file size constraint. Default is 0
		invalid_value : int, optional, default 256
			specifies which is the invalid value used as separator. Default is 256
		"""
		super(CBlackBoxPaddingEvasionProblem, self).__init__(model_wrapper,
															 latent_space_size=how_many_padding_bytes,
															 iterations=iterations,
															 population_size=population_size,
															 is_debug=is_debug,
															 penalty_regularizer=penalty_regularizer)

		self.invalid_value = invalid_value
		self.indexes_to_perturb = list(range(how_many_padding_bytes))

	def apply_feasible_manipulations(self, t, x: CArray) -> CArray:
		"""
		Apply the padding practical manipulation on the input sample
		Parameters
		----------
		t : CArray
			the vector of manipulations in [0,1]
		x : CArray
			the input space sample to perturb

		Returns
		-------
		CArray:
			the adversarial malware
		"""
		byte_values = (t * 255).astype(np.int)
		x_adv = x.append(byte_values)
		return x_adv



import os
import sys

import numpy as np
import torch
import torchvision.transforms as transforms
from secml.array import CArray
from secml.ml.classifiers.pytorch.c_classifier_pytorch import CClassifierPyTorch
from secml.settings import SECML_PYTORCH_USE_CUDA

from secml_malware.models.basee2e import End2EndModel

use_cuda = torch.cuda.is_available() and SECML_PYTORCH_USE_CUDA


class CClassifierEnd2EndMalware(CClassifierPyTorch):

	def __init__(
			self,
			model: End2EndModel,
			epochs=100,
			batch_size=256,
			train_transform=None,
			preprocess=None,
			softmax_outputs=False,
			random_state=None,
			plus_version=False,
			input_shape=(1, 2 ** 20),
			verbose=0,
	):
		super(CClassifierEnd2EndMalware, self).__init__(
			model,
			loss="binary_crossentropy",
			epochs=epochs,
			batch_size=batch_size,
			preprocess=preprocess,
			input_shape=(1, model.max_input_size),
			softmax_outputs=softmax_outputs,
			random_state=random_state,
		)
		self.plus_version = plus_version
		self.verbose = verbose
		self.train_transform = (
			train_transform
			if train_transform is not None
			else transforms.Lambda(lambda p: p.reshape(input_shape[1]))
		)

	def gradient(self, x, w=None):
		"""Compute gradient at x by doing a forward and a backward pass.

		The gradient is pre-multiplied by w.

		"""
		return self._gradient_f(x)

	def gradient_f_x(self, x, **kwargs):
		"""Returns the gradient of the function on point x.
		
		Arguments:
			x {CArray} -- The point
		
		Raises:
			NotImplementedError: Model do not support gradient
		
		Returns:
			CArray -- the gradient computed on x
		"""
		if self.preprocess is not None:
			# Normalize data before compute the classifier gradient
			x_pre = self.preprocess.normalize(x)
		else:  # Data will not be preprocessed
			x_pre = x
		try:  # Get the derivative of decision_function
			grad_f = self._gradient_f(x_pre, **kwargs)
		except NotImplementedError:
			raise NotImplementedError(
				"{:} does not implement `gradient_f_x`".format(self.__class__.__name__)
			)
		return grad_f

	def _gradient_f(self, x, y=None, w=None, layer=None, sum_embedding=True):
		penalty_term = torch.zeros(1)
		penalty_term.requires_grad_()
		gradient = self.compute_embedding_gradient(x.tondarray(), penalty_term)
		if sum_embedding:
			gradient = torch.mean(gradient, dim=1)
		if gradient.is_cuda:
			gradient = gradient.cpu()
		return CArray(gradient)

	def load_pretrained_model(self, path: str = None):
		"""
		Load pretrained model weights

		Parameters
		----------
		path : str, optional, default None
			The path of the model, default is None, and it will load the internal default one
		"""
		root = os.path.dirname(
			os.path.dirname(os.path.abspath(sys.modules["secml_malware"].__file__))
		)
		self._model.load_simplified_model(
			os.path.join(root, "secml_malware/data/trained/pretrained_malconv.pth") if path is None else path)
		self._classes = np.array([0, 1])
		self._n_features = 2 ** 20

	def get_embedding_size(self):
		"""
		Get the embedding space dimensionality

		Returns
		-------
		int
			the dimensionality of the embedding space
		"""
		return self._model.embedding_size

	def get_input_max_length(self):
		"""
		Get the input window length

		Returns
		-------
		int
			the window input length
		"""
		return self._model.max_input_size

	def get_embedding_value(self):
		"""
		Get the value used as padding

		Returns
		-------
		int
			a value that is used for padding the sample
		"""
		return self._model.embedding_value

	def get_is_shifting_values(self):
		"""
		Get if the model shifts the values by one

		Returns
		-------
		bool
			return if the values are shifted by one
		"""
		return self._model.shift_values

	def embed(self, x: CArray, transpose: bool = True):
		"""
		Embed the sample inside the embedding space

		Parameters
		----------
		x : CArray
			the sample to embed
		transpose : bool, optional, default True
			set True to return the transposed feature space vector
		Returns
		-------
		torch.Tensor
			the embedded vector
		"""
		return self._model.embed(x, transpose=transpose)

	def compute_embedding_gradient(self, x: CArray, penalty_term: torch.Tensor):
		"""Compute the gradient w.r.t. embedding layer.
		
		Parameters
		----------
		x : CArray
			point where gradient will be computed
		penalty_term : float
			the penalty term
		
		Returns
		----------
		CArray
			the gradient w.r.t. the embedding
		"""
		data = x
		if isinstance(x, CArray):
			data = x.tondarray()
		emb_x = self.embed(data)
		y = self.model.embedd_and_forward(emb_x)
		output = y + penalty_term
		if use_cuda:
			output = output.cuda()
		g = torch.autograd.grad(output, emb_x)[0]
		g = torch.transpose(g, 1, 2)[0]
		return g

	def embedding_predict(self, x):
		"""
		Embed the sample and produce prediction.

		Parameters
		----------
		x : CArray
			the input sample

		Returns
		-------
		float
			the malware score
		"""
		return self._model.embedd_and_forward(x)

	def _forward(self, x):
		x = x.atleast_2d()
		scores = super(CClassifierEnd2EndMalware, self)._forward(x)
		confidence = []
		for i in range(x.shape[0]):
			confidence.append([1 - scores[i, 1].item(), scores[i, 1].item()])
		confidence = CArray(confidence)
		return confidence



from abc import abstractmethod
from secml.array import CArray
from secml_malware.attack.blackbox.c_wrapper_phi import CWrapperPhi
import numpy as np


def _slice_sequence(slice_size, sequence, irregular=None):
	if irregular is not None:
		expanded_sequence = [
			sequence[: sum(irregular[:i + 1])] for i, _ in enumerate(irregular)
			# sequence[: sum(irregular[:i])] for i, val in enumerate(irregular)
		]
	else:
		how_many = len(sequence) // slice_size
		expanded_sequence = [
			sequence[: (i + 1) * slice_size]
			for i in range(how_many)
		]
	# expanded_sequence = np.array(expanded_sequence)
	return expanded_sequence


class CBlackBoxProblem:
	"""
	Base class for encapsulating a black box optimization problem.
	"""

	def __init__(
			self,
			model_wrapper: CWrapperPhi,
			latent_space_size: int,
			population_size: int,
			penalty_regularizer: float,
			iterations: int,
			seed: int = None,
			is_debug: bool = False,
			hard_label: bool = False,
			threshold: float = 0,
			loss: str = 'l1'
	):
		"""
		Creates an instance of a black-box problem.

		Parameters
		----------
		model_wrapper : CWrapperPhi
			the target model to attack
		latent_space_size : int
			the dimensionality of the vectors of parameters used to perturb malware
		population_size : int
			the population size generated at each round by the genetic algorithm
		penalty_regularizer : float
			the regularization parameter that affects the size constraint
		iterations : int
			the number of iterations performed by the algorithm
		seed : int, optional, default None
			the seed to pass to the genetic engine, leave None for random behaviour.
		is_debug : bool, optional, default False
			if True, it prints debug messages while optimizing
		hard_label : bool, optional default False
			if True, the problem will use only binary labels instead. Infinity will be used for non-evasive samples.
		threshold : float, optional, default 0
			the detection threshold. Leave 0 to test the degradation of the model until the end of the algorithm.
		loss : str, optional, default l1
			The loss function used as objective function
		"""
		self.model_wrapper = model_wrapper
		self.latent_space_size = latent_space_size
		self.population_size = population_size
		self._original_x = None
		self.penalty_regularizer = penalty_regularizer
		self.iterations = iterations
		self.seed = seed
		self.is_debug = is_debug
		self.hard_label = hard_label
		self.threshold = threshold
		self.loss = loss
		self.confidences_ = []
		self.fitness_ = []
		self.sizes_ = []
		self.advx = []

	def clear_results(self):
		"""
		Reset the internal state after computing an attack
		"""
		self.confidences_ = []
		self.fitness_ = []
		self.sizes_ = []
		self.advx = []

	def init_starting_point(self, x: CArray) -> CArray:
		"""
		Initialize the problem, by setting the starting point.

		Parameters
		----------
		x : CArray
			the initial point

		Returns
		-------
		CArray
			the initial point (padded accordingly to remove trailing invalid values)
		"""
		self._original_x = x.deepcopy()
		padding_positions = x.find(x == 256)
		self.clear_results()
		if padding_positions:
			self._original_x = self._original_x[0, :padding_positions[0]]
		return self._original_x

	def get_bounds(self) -> (list, list):
		"""
		Gets the bounds for the genetic algorithm
		Returns
		-------
		list, list
			lower bounds and upper bounds
		"""
		return [0] * self.latent_space_size, [1] * self.latent_space_size

	def fitness(self, t: np.ndarray) -> list:
		"""
		Compute the objective function.

		Parameters
		----------
		t : numpy array
			the vector of parameter to test
		Returns
		-------
		list
			the score attributed by the target combined with the the penalty term
		"""
		candidate = self.apply_feasible_manipulations(t, self._original_x)
		penalty_term = self.compute_penalty_term(self._original_x, candidate, self.penalty_regularizer)
		score = self.score_step(candidate, penalty_term)
		return [score]

	def _compute_loss(self, confidence, penalty):
		if self.loss == 'l1':
			return confidence + penalty
		if self.loss == 'cw':
			return max(confidence - self.threshold + 0.1, 0) + penalty
		if self.loss == 'log':
			return -np.log(1 - confidence) + penalty
		raise ValueError('NO LOSS')

	def score_step(self, x: CArray, penalty_term: float) -> float:
		"""
		Computes the objective function, combining the penalty term

		Parameters
		----------
		x : CArray
			the original sample
		penalty_term : float
			the penalty term

		Returns
		-------
		float, float
		"""
		_, confidence = self.model_wrapper.predict(x, return_decision_function=True)
		confidence = confidence[0, 1].item()
		if self.hard_label:
			confidence = np.infty if confidence > self.threshold else 0
		fitness_value = self._compute_loss(confidence, penalty_term)
		self.confidences_.append(confidence)
		self.sizes_.append(x.shape[-1])
		self.fitness_.append(fitness_value)
		return fitness_value

	def _export_internal_results(self, irregular=None) -> (list, list, list):
		"""
		Exports the results of the attack

		Parameters
		----------
		irregular : list, optional, default None
			Slices the internal results
		Returns
		-------

		"""
		confidence = _slice_sequence(self.population_size, self.confidences_[1:], irregular)
		fitness = _slice_sequence(self.population_size, self.fitness_[1:], irregular)
		sizes = _slice_sequence(self.population_size, self.sizes_[1:], irregular)
		best_idx = [np.argmin(f) for f in fitness]
		fitness = [self.fitness_[0]] + [f[i] for f, i in zip(fitness, best_idx)]
		confidence = [self.confidences_[0]] + [f[i] for f, i in zip(confidence, best_idx)]
		sizes = [self.sizes_[0]] + [f[i] for f, i in zip(sizes, best_idx)]

		return confidence, fitness, sizes

	def compute_penalty_term(self, original_x: CArray, adv_x: CArray, par: float) -> float:
		"""
		Computes the size penalty term

		Parameters
		----------
		original_x : CArray
			the original malware sample
		adv_x : CArray
			the adversarial malware
		par : float
			the regularization parameter

		Returns
		-------
		flaot
			the size penalty term, multiplied by the regularization parameter
		"""
		return par * abs(adv_x.shape[-1] - original_x.shape[-1])

	@abstractmethod
	def apply_feasible_manipulations(self, t: np.ndarray, x: CArray) -> CArray:
		"""
		Applies the manipulation to the malware

		Parameters
		----------
		t : numpy array
			the vector of parameters
		x : CArray
			the original malware sample

		Returns
		-------
		CArray
			the adversarial malware
		"""
		raise NotImplementedError("This method is abstract, you should implement it somewhere else!")



from setuptools import setup, find_packages

with open("README.md", "r") as f:
	long_description = f.read()

setup(
	name="secml_malware",
	version="0.2.5.2",
	author="zangobot",
	author_email="luca.demetrio93@unica.it",
	packages=find_packages(),
	install_requires=[
		'lief',
		'python-magic',
		'matplotlib',
		'numpy',
		'seaborn',
		'secml',
		'lightgbm',
		'deap',
		'torch',
		'torchvision',
		'ember'
	],
	url="https://github.com/pralab/secml_malware",
	license="GPL 3.0",
	short_description="Robustness evaluation and pentesting tool for machine-learning Windows malware detectors",
	long_description=long_description,
	long_description_content_type="text/markdown",
	package_data={'': ['data/trained/pretrained_malconv.pth', 'data/goodware_samples/file*', 'data/malware_samples/test_malware', 'data/malware_samples/test_folder']},
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
		"Operating System :: OS Independent",
	],
	python_requires=">=3.7",
)



# import numpy as np
# from secml.array import CArray
# from secml_malware.attack.whitebox.tests.malware_test_base import End2EndBaseTests
# from secml_malware.visualization.exe import plot_code_segment, plot_header_contribution_histogram,plot_range_histogram, plot_bins
#
#
# class ExeVisualizationTests(End2EndBaseTests):
#
# 	def setUp(self):
# 		super(ExeVisualizationTests, self).setUp()
# 		self.weights = CArray(np.random.randn(2, self.max_length))
# 		self.classifier.load_pretrained_model(self.ember_path)
#
# 	def test_ember_model_histogram(self):
# 		malware = CArray(self.malware)
# 		itgs = CMalConvIntegratedGradients(self.classifier)
# 		attr = itgs.explain(malware, m=250, reference=CArray(self.baseline))
# 		plot_range_histogram(attr, how_many=256, force_plot=False)
#
# 	def test_plot_bins(self):
# 		malware = CArray(self.malware)
# 		cut_index = malware.tolist()[0].index(256)
# 		itgs = CMalConvIntegratedGradients(self.classifier)
# 		attr = itgs.explain(malware, m=40, reference=CArray(self.baseline))
# 		plot_bins(attr, n_bins=2056, cut_index=cut_index, force_plot=False)
#
# 	def test_integrated_gradients_plot(self):
# 		malware = CArray(self.malware)
# 		itgs = CMalConvIntegratedGradients(self.classifier)
# 		attr = itgs.explain(malware, m=50, reference=CArray(self.baseline))
# 		_, conf = self.classifier.predict(malware, return_decision_function=True)
# 		title = 'Confidence: {0:.4f}%\nDOS + COFF + OPT + SECT Headers\nBaseline : empty file'.format(conf[1].item()*100)
# 		plot_code_segment(self.byte_malware, 0, 512, attr, title, force_plot=False, show_positives=True, show_negatives=True)
# 		plot_header_contribution_histogram(self.byte_malware, attr, force_plot=False)
#
# 	def test_code_segment_plot(self):
# 		plot_code_segment(self.byte_malware, 0, 120, self.weights, 'test', show_positives=True, show_negatives=True, force_plot=False)
#
# 	def test_visualize_histogram(self):
# 		plot_header_contribution_histogram(self.byte_malware, self.weights, force_plot=False)



from secml_malware.attack.whitebox.tests.malware_test_base import EmberBaseTests


class EmberTestSuite(EmberBaseTests):
	def test_batch_prediction(self):
		features = self.classifier.extract_features(self.X)
		y_pred = self.classifier.predict(features)

		y_pred1, y_pred2 = self.classifier.predict(features[0, :]), self.classifier.predict(features[1, :])

		self.assert_array_equal(y_pred1, y_pred[0])
		self.assert_array_equal(y_pred2, y_pred[1])



import struct
from copy import copy, deepcopy

from secml.array import CArray
from secml_malware.attack.blackbox.c_blackbox_problem import CBlackBoxProblem
from secml_malware.attack.blackbox.c_wrapper_phi import CWrapperPhi
import numpy as np


class CBlackBoxHeaderEvasionProblem(CBlackBoxProblem):
	"""
	Blackbox attacks that perturbs the DOS header (partially or fully)
	"""
	def __init__(
			self,
			model_wrapper: CWrapperPhi,
			population_size: int,
			optimize_all_dos: bool = False,
			iterations: int = 100,
			is_debug: bool = False,
			penalty_regularizer: float = 0,
			invalid_value: int = 256
	):
		"""
		Creates the attack.

		Parameters
		----------
		model_wrapper : CWrapperPhi
			the target model, wrapped inside a CWrapperPhi
		population_size : int
			the population size generated at each round by the genetic algorithm
		optimize_all_dos : bool
			if True, it manipulates all the DOS header. False to only manipulate only the first 58 bytes
		iterations : int, optional, default 100
			the total number of iterations, default 100
		is_debug : bool, optional, default False
			if True, it prints messages while optimizing. Default is False
		penalty_regularizer : float, optional, default 0
			the penalty regularizer for the file size constraint. Default is 0
		invalid_value : int, optional, default 256
			specifies which is the invalid value used as separator. Default is 256
		"""
		super(CBlackBoxHeaderEvasionProblem, self).__init__(model_wrapper,
															latent_space_size=58,
															iterations=iterations,
															population_size=population_size,
															is_debug=is_debug,
															penalty_regularizer=penalty_regularizer)
		self.optimize_all_dos = optimize_all_dos
		self.invalid_value = invalid_value
		self.indexes_to_perturb = list(range(2, 60))

	def init_starting_point(self, x: CArray) -> CArray:
		"""
		Initialize the problem, by setting the starting point.

		Parameters
		----------
		x : CArray
			the initial point

		Returns
		-------
		CArray
			the initial point (padded accordingly to remove trailing invalid values)
		"""
		self.indexes_to_perturb = list(range(2, 60))
		if self.optimize_all_dos:
			pe_index = struct.unpack('<I', bytes(x[0, 60:64].tolist()[0]))[0]
			self.indexes_to_perturb += list(range(64, pe_index))
		self.latent_space_size = len(self.indexes_to_perturb)
		return super(CBlackBoxHeaderEvasionProblem, self).init_starting_point(x)

	def apply_feasible_manipulations(self, t : np.ndarray, x: CArray) -> CArray:
		"""
		Apply the partial / full DOS practical manipulation

		Parameters
		----------
		t : numpy array
			the vector of manipulations in [0,1]
		x : CArray
			the input space sample to perturb

		Returns
		-------
		CArray:
			the adversarial malware
		"""
		byte_values = (t * 255).astype(np.int)
		x_adv = deepcopy(x)
		x_adv[0, self.indexes_to_perturb] = CArray(byte_values)
		return CArray(x_adv)



from abc import abstractmethod
from os.path import isfile

import numpy as np
import torch
from secml.settings import SECML_PYTORCH_USE_CUDA
from torch.nn import Module

from secml_malware.utils.exceptions import FileNotExistsException

use_cuda = torch.cuda.is_available() and SECML_PYTORCH_USE_CUDA


class End2EndModel(Module):

	def __init__(self, embedding_size: int, max_input_size: int, embedding_value: int, shift_values: bool):
		"""
		Basic end to end model wrapper

		Parameters
		----------
		embedding_size : int
			the size of the embedding space
		max_input_size : int
			the input window size
		embedding_value : int
			the value used as padding
		shift_values : bool
			True if values are shifted by one
		"""
		super(End2EndModel, self).__init__()
		self.embedding_size = embedding_size
		self.max_input_size = max_input_size
		self.embedding_value = embedding_value
		self.shift_values = shift_values

	@classmethod
	def path_to_exe_vector(cls, path: str, max_length: int, padding_value: int, shift_values: bool) -> np.ndarray:
		"""
		Creates a numpy array from the bytes contained in file

		Parameters
		----------
		path : str
			the path of the file
		max_length : int
			the max input size of the network
		padding_value : int
			the value used as padding
		shift_values : bool
			True if values are shifted by one

		Returns
		-------
		numpy array
			the sample as numpy array, cropped or padded
		"""
		exe = cls.load_sample_from_file(path, max_length, has_shape=True, padding_value=padding_value,
										shift_values=shift_values)
		return exe.reshape(max_length)

	@classmethod
	@abstractmethod
	def bytes_to_numpy(cls, bytez: bytes, max_length: int, padding_value: int, shift_values: bool) -> np.ndarray:
		"""
		It creates a numpy array from bare bytes. The vector is max_length long.

		Parameters
		----------
		bytez : bytes
			byte string containing the sample
		max_length : int
			the max input size of the network
		padding_value : int
			the value used as padding
		shift_values : bool
			True if values are shifted by one

		Returns
		-------
		numpy array
			the sample as numpy array, cropped or padded

		"""
		b = np.ones((max_length,), dtype=np.uint16) * padding_value
		bytez = np.frombuffer(bytez[:max_length], dtype=np.uint8)
		bytez = bytez.astype(np.uint16) + shift_values
		b[: len(bytez)] = bytez
		return np.array(b, dtype=float)

	@classmethod
	@abstractmethod
	def list_to_numpy(cls, x, max_length, padding_value, shift_values):
		"""
		It creates a numpy array from bare bytes. The vector is max_length long.
		"""
		b = np.ones((max_length,), dtype=np.uint16) * padding_value
		bytez = np.array(x[:max_length], dtype=np.uint8)
		bytez = bytez.astype(np.uint16) + shift_values
		b[: len(bytez)] = bytez
		return np.array(b, dtype=float)

	@classmethod
	@abstractmethod
	def load_sample_from_file(cls, path, max_length, has_shape, padding_value, shift_values):
		"""
		It creates a numpy array containing a sample. The vector is max_length long.
		If shape is true, then the path is supposed to be a (1,1) matrix.
		Hence, item() is called.
		"""
		file_path = path.item() if has_shape else path
		with open(file_path, "rb") as malware:
			code = cls.bytes_to_numpy(malware.read(), max_length, padding_value, shift_values)
		return code

	@abstractmethod
	def embed(self, input_x, transpose=True):
		"""
		It embeds an input vector into MalConv embedded representation.
		"""
		pass

	# @abstractmethod
	def compute_embedding_gradient(self, numpy_x: np.ndarray) -> torch.Tensor:
		"""
		It computes the gradient w.r.t. the embedding layer.

		Parameters
		----------
		numpy_x : numpy array
			the numpy array containing a sample
		Returns
		-------
		torch.Tensor
			the gradient w.r.t. the embedding layer
		"""
		emb_x = self.embed(numpy_x)
		y = self.embedd_and_forward(emb_x)
		g = torch.autograd.grad(y, emb_x)[0]
		g = torch.transpose(g, 1, 2)[0]
		return g

	@abstractmethod
	def embedd_and_forward(self, x: torch.Tensor) -> torch.Tensor:
		"""
		Compute the embedding for sample x and returns the prediction.

		Parameters
		----------
		x : torch.Tensor
			the sample as torch tensor
		Returns
		-------
		torch.Tensor
			the result of the forward pass
		"""
		pass

	def forward(self, x: torch.Tensor) -> torch.Tensor:
		"""
		Forward pass.

		Parameters
		----------
		x : torch.Tensor
			the sample to test
		Returns
		-------
		torch.Tensor
			the result of the forward pass
		"""
		embedding_1 = self.embed(x.contiguous())
		output = self.embedd_and_forward(embedding_1)
		return output

	def load_simplified_model(self, path: str):
		"""
		Load the model weights.

		Parameters
		----------
		path : str
			the path to the model
		"""
		if not isfile(path):
			raise FileNotExistsException('{} path not pointing to regular file!'.format(path))
		f = torch.load(path) if use_cuda else torch.load(path, map_location="cpu")
		self.load_state_dict(f)
		self.eval()



import unittest

from secml.array import CArray

from secml_malware.attack.whitebox import CHeaderEvasion, CPaddingEvasion, CKreukEvasion, CSuciuEvasion, \
	CContentShiftingEvasion
from secml_malware.attack.whitebox.c_format_exploit_evasion import CFormatExploitEvasion
from secml_malware.attack.whitebox.c_headerfields_evasion import CHeaderFieldsEvasion
from secml_malware.attack.whitebox.tests.malware_test_base import End2EndBaseTests


class EvasionEnd2EndTestSuite(End2EndBaseTests):
	def setUp(self):
		super(EvasionEnd2EndTestSuite, self).setUp()
		self.classifier.load_pretrained_model(self.ember_path)
		self.surrogate_classifier.load_pretrained_model(self.surrogate_path)
		self.Y = self.classifier.predict(CArray(self.X), return_decision_function=False)

	def test_whitebox_ember_model_attack_random_init(self):
		attack = CHeaderEvasion(
			self.classifier,
			is_debug=True,
			random_init=True,
		)
		self.assert_evasion_result(attack)

	def test_whitebox_ember_model_attack_no_random_init(self):
		attack = CHeaderEvasion(
			self.classifier,
			is_debug=True,
			random_init=False,
		)
		y_pred, scores, _, _ = attack.run(self.X, self.Y)
		n_old_y_malw = sum(self.Y == 1)
		n_false_negative = sum(self.Y == 0)
		n_new_detected_malw = sum(y_pred == 1) - n_false_negative
		self.assertNotEqual(
			n_old_y_malw,
			n_new_detected_malw,
			msg="Evasion achieved: {}/{}".format(
				self.Y.shape[0] - n_new_detected_malw, self.Y.shape[0]
			),
		)

	def test_padding_whitebox_ember_model_attack(self):
		padding_attack = CPaddingEvasion(
			self.classifier,
			1000,
			random_init=True,
			is_debug=True
		)
		self.assert_evasion_result(padding_attack)

	def test_kreuk_whitebox_attack(self):
		kreuk_attack = CKreukEvasion(
			self.classifier,
			how_many_padding_bytes=1000,
			epsilon=0.03,
			iterations=50,
			is_debug=True
		)
		self.assert_evasion_result(kreuk_attack)

	def test_kreuk_whitebox_no_slack_attack_p2(self):
		kreuk_attack = CKreukEvasion(
			self.classifier,
			how_many_padding_bytes=1000,
			epsilon=0.03,
			iterations=50,
			is_debug=True,
			compute_slack=False,
			p_norm=2
		)
		self.assert_evasion_result(kreuk_attack)

	def test_kreuk_whitebox_attack_p2(self):
		kreuk_attack = CKreukEvasion(
			self.classifier,
			how_many_padding_bytes=1000,
			epsilon=0.03,
			iterations=50,
			is_debug=True,
			compute_slack=True,
			p_norm=2
		)
		self.assert_evasion_result(kreuk_attack)

	def test_kreuk_whitebox_no_slack_attack(self):
		kreuk_attack = CKreukEvasion(
			self.classifier,
			how_many_padding_bytes=1000,
			epsilon=0.03,
			iterations=50,
			is_debug=True,
			compute_slack=False
		)
		self.assert_evasion_result(kreuk_attack)

	def test_suciu_appending_whitebox_attack(self):
		kreuk_attack = CSuciuEvasion(
			self.classifier,
			how_many_padding_bytes=1000,
			epsilon=0.03,
			is_debug=True
		)
		self.assert_evasion_result(kreuk_attack)

	def test_suciu_appending_whitebox_no_slack_attack(self):
		kreuk_attack = CSuciuEvasion(
			self.classifier,
			how_many_padding_bytes=1000,
			epsilon=0.03,
			is_debug=True,
			compute_slack=False
		)
		self.assert_evasion_result(kreuk_attack)

	def test_section_shift_attack(self):
		shift_attack = CContentShiftingEvasion(
			self.classifier,
			preferable_extension_amount=0x200,
			iterations=2,
			is_debug=True
		)
		self.assert_evasion_result(shift_attack)

	def test_pe_shift_attack(self):
		shift_attack = CFormatExploitEvasion(
			self.classifier,
			preferable_extension_amount=0,
			pe_header_extension=0x200,
			iterations=20,
			is_debug=True
		)
		self.assert_evasion_result(shift_attack)

	def test_header_fields_attack(self):
		header_fields_attack = CHeaderFieldsEvasion(
			self.classifier,
			iterations=20,
			is_debug=True
		)
		self.assert_evasion_result(header_fields_attack)

	def assert_evasion_result(self, attack, create_adv_malware=False):
		y_pred, _, adv_ds, _ = attack.run(self.X, self.Y)
		n_old_y_malw = sum(self.Y == 1)
		n_false_negative = sum(self.Y == 0)
		n_new_detected_malw = sum(y_pred == 1) - n_false_negative
		if create_adv_malware:
			for i, f in enumerate(self.complete_malware_paths):
				attack.create_real_sample_from_adv(f, adv_ds.X[i, :], f + '.adv')
		self.assertNotEqual(
			n_old_y_malw,
			n_new_detected_malw,
			msg="Evasion achieved: {}/{}".format(
				self.Y.shape[0] - n_new_detected_malw, self.Y.shape[0]
			),
		)


if __name__ == "__main__":
	unittest.main()
